""" Click commands for ``bob.measure`` """

from .. import load
from . import common_options, figure

SCORE_FORMAT = (
    "The command takes as input generic 2-column data format as "
    "specified in the documentation of "
    ":py:func:`bob.measure.load.split`."
)
CRITERIA = ("eer", "min-hter", "far")


@common_options.metrics_command(
    common_options.METRICS_HELP.format(
        names="FPR, FNR, precision, recall, F1-score, AUC ROC",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        hter_note=" ",
        command="bob measure metrics",
    ),
    criteria=CRITERIA,
    far_name="FPR",
)
def metrics(ctx, scores, evaluation, **kwargs):
    process = figure.Metrics(ctx, scores, evaluation, load.split)
    process.run()


@common_options.roc_command(
    common_options.ROC_HELP.format(
        score_format=SCORE_FORMAT, command="bob measure roc"
    ),
    far_name="FPR",
)
def roc(ctx, scores, evaluation, **kwargs):
    process = figure.Roc(ctx, scores, evaluation, load.split)
    process.run()


@common_options.det_command(
    common_options.DET_HELP.format(
        score_format=SCORE_FORMAT, command="bob measure det"
    ),
    far_name="FPR",
)
def det(ctx, scores, evaluation, **kwargs):
    process = figure.Det(ctx, scores, evaluation, load.split)
    process.run()


@common_options.epc_command(
    common_options.EPC_HELP.format(
        score_format=SCORE_FORMAT, command="bob measure epc"
    )
)
def epc(ctx, scores, **kwargs):
    process = figure.Epc(ctx, scores, True, load.split)
    process.run()


@common_options.hist_command(
    common_options.HIST_HELP.format(
        score_format=SCORE_FORMAT, command="bob measure hist"
    ),
    far_name="FPR",
)
def hist(ctx, scores, evaluation, **kwargs):
    process = figure.Hist(ctx, scores, evaluation, load.split)
    process.run()


@common_options.evaluate_command(
    common_options.EVALUATE_HELP.format(
        score_format=SCORE_FORMAT, command="bob measure evaluate"
    ),
    criteria=CRITERIA,
    far_name="FPR",
)
def evaluate(ctx, scores, evaluation, **kwargs):
    common_options.evaluate_flow(
        ctx, scores, evaluation, metrics, roc, det, epc, hist, **kwargs
    )


@common_options.multi_metrics_command(
    common_options.MULTI_METRICS_HELP.format(
        names="FtA, FAR, FRR, FMR, FMNR, HTER",
        criteria=CRITERIA,
        score_format=SCORE_FORMAT,
        command="bob measure multi-metrics",
    ),
    criteria=CRITERIA,
    far_name="FPR",
)
def multi_metrics(ctx, scores, evaluation, protocols_number, **kwargs):
    ctx.meta["min_arg"] = protocols_number * (2 if evaluation else 1)
    process = figure.MultiMetrics(ctx, scores, evaluation, load.split)
    process.run()



"""The main entry for bob.measure (click-based) scripts.
"""
import click
import pkg_resources

from clapper.click import AliasedGroup
from click_plugins import with_plugins


@with_plugins(pkg_resources.iter_entry_points("bob.measure.cli"))
@click.group(cls=AliasedGroup)
def measure():
    """Generic performance evaluation commands."""
    pass



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Mon 23 May 2011 14:36:14 CEST
import warnings

import numpy


def log_values(min_step=-4, counts_per_step=4):
    """Computes log-scaled values between :math:`10^{M}` and 1

    This function computes log-scaled values between :math:`10^{M}` and 1
    (including), where :math:`M` is the ``min_ste`` argument, which needs to be a
    negative integer.  The integral ``counts_per_step`` value defines how many
    values between two adjacent powers of 10 will be created.  The total number
    of values will be ``-min_step * counts_per_step + 1``.


    Parameters:

      min_step (:py:class:`int`, optional): The power of 10 that will be the
        minimum value.  E.g., the default ``-4`` will result in the first number
        to be :math:`10^{-4}` = ``0.00001`` or ``0.01%``

      counts_per_step (:py:class:`int`, optional): The number of values that will
        be put between two adjacent powers of 10.  With the default value ``4``
        (and default values of ``min_step``), we will get ``log_list[0] ==
        1e-4``, ``log_list[4] == 1e-3``, ..., ``log_list[16] == 1``.


    Returns:

      :py:class:`list`: A list of logarithmically scaled values between
      :math:`10^{M}` and 1.

    """

    import math

    return [
        math.pow(10.0, i * 1.0 / counts_per_step)
        for i in range(min_step * counts_per_step, 0)
    ] + [1.0]


def _semilogx(x, y, **kwargs):
    # remove points were x is 0
    x, y = numpy.asarray(x), numpy.asarray(y)
    zero_index = x == 0
    x = x[~zero_index]
    y = y[~zero_index]
    from matplotlib import pyplot

    return pyplot.semilogx(x, y, **kwargs)


def roc(
    negatives,
    positives,
    npoints=2000,
    CAR=None,
    min_far=-8,
    tpr=False,
    semilogx=False,
    **kwargs,
):
    """Plots Receiver Operating Characteristic (ROC) curve.

    This method will call ``matplotlib`` to plot the ROC curve for a system which
    contains a particular set of negatives (impostors) and positives (clients)
    scores. We use the standard :py:func:`matplotlib.pyplot.plot` command. All
    parameters passed with exception of the three first parameters of this method
    will be directly passed to the plot command.

    The plot will represent the false-alarm on the horizontal axis and the
    false-rejection on the vertical axis.  The values for the axis will be
    computed using :py:func:`bob.measure.roc`.

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting command. You are the responsible for setting up and
      saving the figure as you see fit.


    Parameters
    ----------
    negatives : array
        1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier. See
        (:py:func:`bob.measure.roc`)

    positives : array
        1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier. See
        (:py:func:`bob.measure.roc`)

    npoints : :py:class:`int`, optional
        The number of points for the plot. See
        (:py:func:`bob.measure.roc`)

    min_far : float, optional
        The minimum value of FPR and FNR that is used for ROC computations.

    tpr : bool, optional
        If True, will plot TPR (TPR = 1 - FNR) on the y-axis instead of FNR.

    semilogx : bool, optional
        If True, will use pyplot.semilogx to plot the ROC curve.

    CAR : :py:class:`bool`, optional
        This option is deprecated. Please use ``TPR`` and ``semilogx`` options instead.
        If set to ``True``, it will plot the CPR
        (CAR) over FPR in using :py:func:`matplotlib.pyplot.semilogx`, otherwise the
        FPR over FNR linearly using :py:func:`matplotlib.pyplot.plot`.

    **kwargs
        Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.

    Returns
    -------
    object
        `list` of :py:class:`matplotlib.lines.Line2D`: The lines that
        were added as defined by the return value of
        :py:func`matplotlib.pyplot.plot`.
    """
    if CAR is not None:
        warnings.warn(
            "CAR argument is deprecated. Please use TPR and semilogx arguments instead.",
            DeprecationWarning,
        )
        tpr = semilogx = CAR

    from matplotlib import pyplot

    from . import roc as calc

    fpr, fnr = calc(negatives, positives, npoints, min_far=min_far)

    if tpr:
        fnr = 1 - fnr  # plot tpr instead of fnr

    if not semilogx:
        return pyplot.plot(fpr, fnr, **kwargs)
    else:
        return _semilogx(fpr, fnr, **kwargs)


def roc_for_far(
    negatives, positives, far_values=log_values(), CAR=True, **kwargs
):
    """Plots the ROC curve for the given list of False Positive Rates (FAR).

    This method will call ``matplotlib`` to plot the ROC curve for a system which
    contains a particular set of negatives (impostors) and positives (clients)
    scores. We use the standard :py:func:`matplotlib.pyplot.semilogx` command.
    All parameters passed with exception of the three first parameters of this
    method will be directly passed to the plot command.

    The plot will represent the False Positive Rate (FPR) on the horizontal
    axis and the Correct Positive Rate (CPR) on the vertical axis.  The values
    for the axis will be computed using :py:func:`bob.measure.roc_for_far`.

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting command. You are the responsible for setting up and
      saving the figure as you see fit.


    Parameters:

      negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier. See
        (:py:func:`bob.measure.roc`)

      positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier. See
        (:py:func:`bob.measure.roc`)

      far_values (:py:class:`list`, optional): The values for the FPR, where the
        CPR (CAR) should be plotted; each value should be in range [0,1].

      CAR (:py:class:`bool`, optional): If set to ``True``, it will plot the CPR
        (CAR) over FPR in using :py:func:`matplotlib.pyplot.semilogx`, otherwise the
        FPR over FNR linearly using :py:func:`matplotlib.pyplot.plot`.

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      :py:class:`list` of :py:class:`matplotlib.lines.Line2D`: The lines that
      were added as defined by the return value of
      :py:func:`matplotlib.pyplot.semilogx`.

    """
    warnings.warn(
        "roc_for_far is deprecated. Please use the roc function instead."
    )

    from matplotlib import pyplot

    from . import roc_for_far as calc

    out = calc(negatives, positives, far_values)
    if not CAR:
        return pyplot.plot(out[0, :], out[1, :], **kwargs)
    else:
        return _semilogx(out[0, :], (1 - out[1, :]), **kwargs)


def precision_recall_curve(negatives, positives, npoints=2000, **kwargs):
    """Plots a Precision-Recall curve.

    This method will call ``matplotlib`` to plot the precision-recall curve for a
    system which contains a particular set of ``negatives`` (impostors) and
    ``positives`` (clients) scores. We use the standard
    :py:func:`matplotlib.pyplot.plot` command. All parameters passed with
    exception of the three first parameters of this method will be directly
    passed to the plot command.

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting command. You are the responsible for setting up and
      saving the figure as you see fit.


    Parameters:

      negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier. See
        (:py:func:`bob.measure.precision_recall_curve`)

      positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier. See
        (:py:func:`bob.measure.precision_recall_curve`)

      npoints (:py:class:`int`, optional): The number of points for the plot. See
        (:py:func:`bob.measure.precision_recall_curve`)

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      :py:class:`list` of :py:class:`matplotlib.lines.Line2D`: The lines that
      were added as defined by the return value of
      :py:func:`matplotlib.pyplot.plot`.

    """

    from matplotlib import pyplot

    from . import precision_recall_curve as calc

    out = calc(negatives, positives, npoints)
    return pyplot.plot(100.0 * out[0, :], 100.0 * out[1, :], **kwargs)


def epc(
    dev_negatives,
    dev_positives,
    test_negatives,
    test_positives,
    npoints=100,
    **kwargs,
):
    """Plots Expected Performance Curve (EPC) as defined in the paper:

    Bengio, S., Keller, M., Mariéthoz, J. (2004). The Expected Performance Curve.
    International Conference on Machine Learning ICML Workshop on ROC Analysis in
    Machine Learning, 136(1), 1963–1966. IDIAP RR. Available:
    http://eprints.pascal-network.org/archive/00000670/

    This method will call ``matplotlib`` to plot the EPC curve for a system which
    contains a particular set of negatives (impostors) and positives (clients)
    for both the development and test sets. We use the standard
    :py:func:`matplotlib.pyplot.plot` command. All parameters passed with
    exception of the five first parameters of this method will be directly passed
    to the plot command.

    The plot will represent the minimum HTER on the vertical axis and the cost on
    the horizontal axis.

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting commands. You are the responsible for setting up and
      saving the figure as you see fit.


    Parameters:

      dev_negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier, from the
        development set. See (:py:func:`bob.measure.epc`)

      dev_positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier, from the
        development set. See (:py:func:`bob.measure.epc`)

      test_negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier, from the test
        set. See (:py:func:`bob.measure.epc`)

      test_positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier, from the test set.
        See (:py:func:`bob.measure.epc`)

      npoints (:py:class:`int`, optional): The number of points for the plot. See
        (:py:func:`bob.measure.epc`)

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      :py:class:`list` of :py:class:`matplotlib.lines.Line2D`: The lines that
      were added as defined by the return value of
      :py:func:`matplotlib.pyplot.plot`.

    """

    from matplotlib import pyplot

    from . import epc as calc

    out = calc(
        dev_negatives, dev_positives, test_negatives, test_positives, npoints
    )
    return pyplot.plot(out[0, :], 100.0 * out[1, :], **kwargs)


def det(negatives, positives, npoints=2000, min_far=-8, **kwargs):
    """Plots Detection Error Trade-off (DET) curve as defined in the paper:

    Martin, A., Doddington, G., Kamm, T., Ordowski, M., & Przybocki, M. (1997).
    The DET curve in assessment of detection task performance. Fifth European
    Conference on Speech Communication and Technology (pp. 1895-1898). Available:
    http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.117.4489&rep=rep1&type=pdf

    This method will call ``matplotlib`` to plot the DET curve(s) for a system
    which contains a particular set of negatives (impostors) and positives
    (clients) scores. We use the standard :py:func:`matplotlib.pyplot.plot`
    command. All parameters passed with exception of the three first parameters
    of this method will be directly passed to the plot command.

    The plot will represent the false-alarm on the horizontal axis and the
    false-rejection on the vertical axis.

    This method is strongly inspired by the NIST implementation for Matlab,
    called DETware, version 2.1 and available for download at the NIST website:

    http://www.itl.nist.gov/iad/mig/tools/

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting commands. You are the responsible for setting up and
      saving the figure as you see fit.

    .. note::

      If you wish to reset axis zooming, you must use the Gaussian scale rather
      than the visual marks showed at the plot, which are just there for
      displaying purposes. The real axis scale is based on
      :py:func:`bob.measure.ppndf`.  For example, if you wish to set the x and y
      axis to display data between 1% and 40% here is the recipe:

      .. code-block:: python

        import bob.measure
        from matplotlib import pyplot
        bob.measure.plot.det(...) #call this as many times as you need
        #AFTER you plot the DET curve, just set the axis in this way:
        pyplot.axis([bob.measure.ppndf(k/100.0) for k in (1, 40, 1, 40)])

      We provide a convenient way for you to do the above in this module. So,
      optionally, you may use the :py:func:`bob.measure.plot.det_axis` method
      like this:

      .. code-block:: python

        import bob.measure
        bob.measure.plot.det(...)
        # please note we convert percentage values in det_axis()
        bob.measure.plot.det_axis([1, 40, 1, 40])


    Parameters:

      negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier. See
        (:py:func:`bob.measure.det`)

      positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier. See
        (:py:func:`bob.measure.det`)

      npoints (:py:class:`int`, optional): The number of points for the plot. See
        (:py:func:`bob.measure.det`)

      axisfontsize (:py:class:`str`, optional): The size to be used by
        x/y-tick-labels to set the font size on the axis

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      :py:class:`list` of :py:class:`matplotlib.lines.Line2D`: The lines that
      were added as defined by the return value of
      :py:func:`matplotlib.pyplot.plot`.

    """

    # these are some constants required in this method
    desiredTicks = [
        "0.000001",
        "0.000002",
        "0.000005",
        "0.00001",
        "0.00002",
        "0.00005",
        "0.0001",
        "0.0002",
        "0.0005",
        "0.001",
        "0.002",
        "0.005",
        "0.01",
        "0.02",
        "0.05",
        "0.1",
        "0.2",
        "0.4",
        "0.6",
        "0.8",
        "0.9",
        "0.95",
        "0.98",
        "0.99",
        "0.995",
        "0.998",
        "0.999",
        "0.9995",
        "0.9998",
        "0.9999",
        "0.99995",
        "0.99998",
        "0.99999",
    ]

    desiredLabels = [
        "0.0001",
        "0.0002",
        "0.0005",
        "0.001",
        "0.002",
        "0.005",
        "0.01",
        "0.02",
        "0.05",
        "0.1",
        "0.2",
        "0.5",
        "1",
        "2",
        "5",
        "10",
        "20",
        "40",
        "60",
        "80",
        "90",
        "95",
        "98",
        "99",
        "99.5",
        "99.8",
        "99.9",
        "99.95",
        "99.98",
        "99.99",
        "99.995",
        "99.998",
        "99.999",
    ]

    # this will actually do the plotting
    from matplotlib import pyplot

    from . import det as calc
    from . import ppndf

    out = calc(negatives, positives, npoints, min_far)
    retval = pyplot.plot(out[0, :], out[1, :], **kwargs)

    # now the trick: we must plot the tick marks by hand using the PPNDF method
    pticks = ppndf(numpy.array(desiredTicks, dtype=float))
    ax = pyplot.gca()  # and finally we set our own tick marks
    ax.set_xticks(pticks)
    ax.set_xticklabels(desiredLabels)
    ax.set_yticks(pticks)
    ax.set_yticklabels(desiredLabels)

    return retval


def det_axis(v, **kwargs):
    """Sets the axis in a DET plot.

    This method wraps the :py:func:`matplotlib.pyplot.axis` by calling
    :py:func:`bob.measure.ppndf` on the values passed by the user so they are
    meaningful in a DET plot as performed by :py:func:`bob.measure.plot.det`.


    Parameters:

      v (``sequence``): A sequence (list, tuple, array or the like) containing
        the X and Y limits in the order ``(xmin, xmax, ymin, ymax)``. Expected
        values should be in percentage (between 0 and 100%).  If ``v`` is not a
        list or tuple that contains 4 numbers it is passed without further
        inspection to :py:func:`matplotlib.pyplot.axis`.

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.axis`.


    Returns:

      object: Whatever is returned by :py:func:`matplotlib.pyplot.axis`.

    """

    import logging

    logger = logging.getLogger("bob.measure")

    from matplotlib import pyplot

    from . import ppndf

    # treat input
    try:
        tv = list(v)  # normal input
        if len(tv) != 4:
            raise IndexError
        tv = [ppndf(float(k) / 100) for k in tv]
        cur = pyplot.axis()

        # limits must be within bounds
        if tv[0] < cur[0]:
            logger.warn("Readjusting xmin: the provided value is out of bounds")
            tv[0] = cur[0]
        if tv[1] > cur[1]:
            logger.warn("Readjusting xmax: the provided value is out of bounds")
            tv[1] = cur[1]
        if tv[2] < cur[2]:
            logger.warn("Readjusting ymin: the provided value is out of bounds")
            tv[2] = cur[2]
        if tv[3] > cur[3]:
            logger.warn("Readjusting ymax: the provided value is out of bounds")
            tv[3] = cur[3]

    except Exception:
        tv = v

    return pyplot.axis(tv, **kwargs)


def cmc(cmc_scores, logx=True, **kwargs):
    """Plots the (cumulative) match characteristics and returns the maximum rank.

    This function plots a CMC curve using the given CMC scores (:py:class:`list`:
        A list of tuples, where each tuple contains the
        ``negative`` and ``positive`` scores for one probe of the database).


    Parameters:

      cmc_scores (array): 1D float array containing the CMC values (See
        :py:func:`bob.measure.cmc`)

      logx (:py:class:`bool`, optional): If set (the default), plots the rank
        axis in logarithmic scale using :py:func:`matplotlib.pyplot.semilogx` or
        in linear scale using :py:func:`matplotlib.pyplot.plot`

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      int: The number of classes (clients) in the given scores.

    """

    from matplotlib import pyplot

    from . import cmc as calc

    out = calc(cmc_scores)

    if logx:
        _semilogx(range(1, len(out) + 1), out, **kwargs)
    else:
        pyplot.plot(range(1, len(out) + 1), out, **kwargs)

    return len(out)


def detection_identification_curve(
    cmc_scores, far_values=log_values(), rank=1, logx=True, **kwargs
):
    """Plots the Detection & Identification curve over the FPR

    This curve is designed to be used in an open set identification protocol, and
    defined in Chapter 14.1 of [LiJain2005]_.  It requires to have at least one
    open set probe item, i.e., with no corresponding gallery, such that the
    positives for that pair are ``None``.

    The detection and identification curve first computes FPR thresholds based on
    the out-of-set probe scores (negative scores).  For each probe item, the
    **maximum** negative score is used.  Then, it plots the detection and
    identification rates for those thresholds, which are based on the in-set
    probe scores only. See [LiJain2005]_ for more details.

    .. [LiJain2005] **Stan Li and Anil K. Jain**, *Handbook of Face Recognition*, Springer, 2005


    Parameters:

      cmc_scores (array): 1D float array containing the CMC values (See
        :py:func:`bob.measure.cmc`)

      rank (:py:class:`int`, optional): The rank for which the curve should be
        plotted

      far_values (:py:class:`list`, optional): The values for the FPR (FAR), where the
        CPR (CAR) should be plotted; each value should be in range [0,1].

      logx (:py:class:`bool`, optional): If set (the default), plots the rank
        axis in logarithmic scale using :py:func:`matplotlib.pyplot.semilogx` or
        in linear scale using :py:func:`matplotlib.pyplot.plot`

      kwargs (:py:class:`dict`, optional): Extra plotting parameters, which are
        passed directly to :py:func:`matplotlib.pyplot.plot`.


    Returns:

      :py:class:`list` of :py:class:`matplotlib.lines.Line2D`: The lines that
      were added as defined by the return value of
      :py:func:`matplotlib.pyplot.plot`.

    """

    import math

    import numpy

    from matplotlib import pyplot

    from . import detection_identification_rate, far_threshold

    # for each probe, for which no positives exists, get the highest negative
    # score; and sort them to compute the FAR thresholds
    negatives = sorted(
        max(neg)
        for neg, pos in cmc_scores
        if (pos is None or not numpy.array(pos).size) and neg is not None
    )
    if not negatives:
        raise ValueError(
            "There need to be at least one pair with only negative scores"
        )

    # compute thresholds based on FAR values
    thresholds = [far_threshold(negatives, [], v, True) for v in far_values]

    # compute detection and identification rate based on the thresholds for
    # the given rank
    rates = [
        detection_identification_rate(cmc_scores, t, rank)
        if not math.isnan(t)
        else numpy.nan
        for t in thresholds
    ]

    # plot curve
    if logx:
        return _semilogx(far_values, rates, **kwargs)
    else:
        return pyplot.plot(far_values, rates, **kwargs)



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Mon 23 May 2011 16:23:05 CEST

"""A set of utilities to load score files with different formats.
"""

import logging

import numpy

LOGGER = logging.getLogger("bob.measure")


def split(filename):
    """split(filename) -> negatives, positives

    Loads the scores from the given file and splits them into positive
    and negative arrays. The file must be a two columns file where the first
    column contains -1 or 1 (for negative or positive respectively) and the
    second the corresponding scores.

    Parameters
    ----------
    filename: :py:class:`str`:
        The name of the file containing the scores.

    Returns
    -------
    negatives: 1D :py:class:`numpy.ndarray` of type float
        This array contains the list of negative scores

    positives: 1D :py:class:`numpy.ndarray` of type float
        This array contains the list of positive scores

    """
    try:
        columns = numpy.loadtxt(filename)
        neg_pos = columns[:, 0]
        scores = columns[:, 1]
    except Exception:
        LOGGER.error(
            """Cannot read {}. This file must be a two columns file with
                   the first column containing -1 or 1 (i.e. negative or
                   positive) and the second the scores
                     (float).""".format(
                filename
            )
        )
        raise
    positives = neg_pos == 1
    return (scores[~positives], scores[positives])


def split_files(filenames):
    """split_files

    Parse a list of files using :py:func:`split`

    Parameters
    ----------

    filenames :
        :any:`list`: A list of file paths

    Returns
    -------
    :any:`list`: A list of tuples, where each tuple contains the
    ``negative`` and ``positive`` scores for one probe of the database. Both
    ``negatives`` and ``positives`` can be either an 1D
    :py:class:`numpy.ndarray` of type ``float``, or ``None``.
    """
    if filenames is None:
        return None
    res = []
    for file_path in filenames:
        try:
            res.append(split(file_path))
        except Exception:
            raise
    return res



"""Stores click common options for plots"""

import functools
import logging

import click
import matplotlib.pyplot as plt
import tabulate

from clapper.click import verbosity_option
from click.types import FLOAT, INT
from matplotlib.backends.backend_pdf import PdfPages

LOGGER = logging.getLogger(__name__)


def bool_option(name, short_name, desc, dflt=False, **kwargs):
    """Generic provider for boolean options

    Parameters
    ----------
    name : str
        name of the option
    short_name : str
        short name for the option
    desc : str
        short description for the option
    dflt : bool or None
        Default value
    **kwargs
        All kwargs are passed to click.option.

    Returns
    -------
    ``callable``
        A decorator to be used for adding this option.
    """

    def custom_bool_option(func):
        def callback(ctx, param, value):
            ctx.meta[name.replace("-", "_")] = value
            return value

        return click.option(
            "-%s/-n%s" % (short_name, short_name),
            "--%s/--no-%s" % (name, name),
            default=dflt,
            help=desc,
            show_default=True,
            callback=callback,
            is_eager=True,
            **kwargs,
        )(func)

    return custom_bool_option


def list_float_option(name, short_name, desc, nitems=None, dflt=None, **kwargs):
    """Get option to get a list of float f

    Parameters
    ----------
    name : str
        name of the option
    short_name : str
        short name for the option
    desc : str
        short description for the option
    nitems : obj:`int`, optional
        If given, the parsed list must contains this number of items.
    dflt : :any:`list`, optional
        List of default  values for axes.
    **kwargs
        All kwargs are passed to click.option.

    Returns
    -------
    ``callable``
        A decorator to be used for adding this option.
    """

    def custom_list_float_option(func):
        def callback(ctx, param, value):
            if value is None or not value.replace(" ", ""):
                value = None
            elif value is not None:
                tmp = value.split(",")
                if nitems is not None and len(tmp) != nitems:
                    raise click.BadParameter(
                        "%s Must provide %d axis limits" % (name, nitems)
                    )
                try:
                    value = [float(i) for i in tmp]
                except Exception:
                    raise click.BadParameter("Inputs of %s be floats" % name)
            ctx.meta[name.replace("-", "_")] = value
            return value

        return click.option(
            "-" + short_name,
            "--" + name,
            default=dflt,
            show_default=True,
            help=desc + " Provide just a space (' ') to cancel default values.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_list_float_option


def open_file_mode_option(**kwargs):
    """Get open mode file option

    Parameters
    ----------
    **kwargs
        All kwargs are passed to click.option.

    Returns
    -------
    ``callable``
        A decorator to be used for adding this option.
    """

    def custom_open_file_mode_option(func):
        def callback(ctx, param, value):
            if value not in ["w", "a", "w+", "a+"]:
                raise click.BadParameter("Incorrect open file mode")
            ctx.meta["open_mode"] = value
            return value

        return click.option(
            "-om",
            "--open-mode",
            default="w",
            help="File open mode",
            callback=callback,
            **kwargs,
        )(func)

    return custom_open_file_mode_option


def scores_argument(min_arg=1, force_eval=False, **kwargs):
    """Get the argument for scores, and add `dev-scores` and `eval-scores` in
    the context when `--eval` flag is on (default)

    Parameters
    ----------
    min_arg : int
        the minimum number of file needed to evaluate a system. For example,
        vulnerability analysis needs licit and spoof and therefore min_arg = 2

    Returns
    -------
     callable
      A decorator to be used for adding score arguments for click commands
    """

    def custom_scores_argument(func):
        def callback(ctx, param, value):
            min_a = min_arg or 1
            mutli = 1
            error = ""
            if (
                "evaluation" in ctx.meta and ctx.meta["evaluation"]
            ) or force_eval:
                mutli += 1
                error += "- %d evaluation file(s) \n" % min_a
            if "train" in ctx.meta and ctx.meta["train"]:
                mutli += 1
                error += "- %d training file(s) \n" % min_a
            # add more test here if other inputs are needed

            min_a *= mutli
            ctx.meta["min_arg"] = min_a
            if len(value) < 1 or len(value) % ctx.meta["min_arg"] != 0:
                raise click.BadParameter(
                    "The number of provided scores must be > 0 and a multiple of %d "
                    "because the following files are required:\n"
                    "- %d development file(s)\n" % (min_a, min_arg or 1)
                    + error,
                    ctx=ctx,
                )
            ctx.meta["scores"] = value
            return value

        return click.argument(
            "scores", type=click.Path(exists=True), callback=callback, **kwargs
        )(func)

    return custom_scores_argument


def alpha_option(dflt=1, **kwargs):
    """An alpha option for plots"""

    def custom_eval_option(func):
        def callback(ctx, param, value):
            ctx.meta["alpha"] = value
            return value

        return click.option(
            "-a",
            "--alpha",
            default=dflt,
            type=click.FLOAT,
            show_default=True,
            help="Adjusts transparency of plots.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_eval_option


def no_legend_option(dflt=True, **kwargs):
    """Get option flag to say if legend should be displayed or not"""
    return bool_option(
        "disp-legend", "dl", "If set, no legend will be printed.", dflt=dflt
    )


def eval_option(**kwargs):
    """Get option flag to say if eval-scores are provided"""

    def custom_eval_option(func):
        def callback(ctx, param, value):
            ctx.meta["evaluation"] = value
            return value

        return click.option(
            "-e",
            "--eval",
            "evaluation",
            is_flag=True,
            default=False,
            show_default=True,
            help="If set, evaluation scores must be provided",
            callback=callback,
            **kwargs,
        )(func)

    return custom_eval_option


def hide_dev_option(dflt=False, **kwargs):
    """Get option flag to say if dev plot should be hidden"""

    def custom_hide_dev_option(func):
        def callback(ctx, param, value):
            ctx.meta["hide_dev"] = value
            return value

        return click.option(
            "--hide-dev",
            is_flag=True,
            default=dflt,
            show_default=True,
            help="If set, hide dev related plots",
            callback=callback,
            **kwargs,
        )(func)

    return custom_hide_dev_option


def sep_dev_eval_option(dflt=True, **kwargs):
    """Get option flag to say if dev and eval plots should be in different
    plots"""
    return bool_option(
        "split",
        "s",
        "If set, evaluation and dev curve in different plots",
        dflt,
    )


def linestyles_option(dflt=False, **kwargs):
    """Get option flag to turn on/off linestyles"""
    return bool_option(
        "line-styles",
        "S",
        "If given, applies a different line style to each line.",
        dflt,
        **kwargs,
    )


def cmc_option(**kwargs):
    """Get option flag to say if cmc scores"""
    return bool_option(
        "cmc", "C", "If set, CMC score files are provided", **kwargs
    )


def semilogx_option(dflt=False, **kwargs):
    """Option to use semilog X-axis"""
    return bool_option(
        "semilogx", "G", "If set, use semilog on X axis", dflt, **kwargs
    )


def tpr_option(dflt=False, **kwargs):
    """Option to use TPR (true positive rate) on y-axis"""
    return bool_option(
        "tpr",
        "tpr",
        "If set, use TPR (also called 1-FNR, 1-FNMR, or 1-BPCER) on Y axis",
        dflt,
        **kwargs,
    )


def print_filenames_option(dflt=True, **kwargs):
    """Option to tell if filenames should be in the title"""
    return bool_option(
        "show-fn", "P", "If set, show filenames in title", dflt, **kwargs
    )


def const_layout_option(dflt=True, **kwargs):
    """Option to set matplotlib constrained_layout"""

    def custom_layout_option(func):
        def callback(ctx, param, value):
            ctx.meta["clayout"] = value
            plt.rcParams["figure.constrained_layout.use"] = value
            return value

        return click.option(
            "-Y/-nY",
            "--clayout/--no-clayout",
            default=dflt,
            show_default=True,
            help="(De)Activate constrained layout",
            callback=callback,
            **kwargs,
        )(func)

    return custom_layout_option


def axes_val_option(dflt=None, **kwargs):
    """Option for setting min/max values on axes"""
    return list_float_option(
        name="axlim",
        short_name="L",
        desc="min/max axes values separated by commas (e.g. ``--axlim "
        " 0.1,100,0.1,100``)",
        nitems=4,
        dflt=dflt,
        **kwargs,
    )


def thresholds_option(**kwargs):
    """Option to give a list of thresholds"""
    return list_float_option(
        name="thres",
        short_name="T",
        desc="Given threshold for metrics computations, e.g. "
        "0.005,0.001,0.056",
        nitems=None,
        dflt=None,
        **kwargs,
    )


def lines_at_option(dflt="1e-3", **kwargs):
    """Get option to draw const far line"""
    return list_float_option(
        name="lines-at",
        short_name="la",
        desc="If given, draw vertical lines at the given axis positions. "
        "Your values must be separated with a comma (,) without space. "
        "This option works in ROC and DET curves.",
        nitems=None,
        dflt=dflt,
        **kwargs,
    )


def x_rotation_option(dflt=0, **kwargs):
    """Get option for rotartion of the x axis lables"""

    def custom_x_rotation_option(func):
        def callback(ctx, param, value):
            value = abs(value)
            ctx.meta["x_rotation"] = value
            return value

        return click.option(
            "-r",
            "--x-rotation",
            type=click.INT,
            default=dflt,
            show_default=True,
            help="X axis labels ration",
            callback=callback,
            **kwargs,
        )(func)

    return custom_x_rotation_option


def legend_ncols_option(dflt=3, **kwargs):
    """Get option for number of columns for legends"""

    def custom_legend_ncols_option(func):
        def callback(ctx, param, value):
            value = abs(value)
            ctx.meta["legends_ncol"] = value
            return value

        return click.option(
            "-lc",
            "--legends-ncol",
            type=click.INT,
            default=dflt,
            show_default=True,
            help="The number of columns of the legend layout.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_legend_ncols_option


def subplot_option(dflt=111, **kwargs):
    """Get option to set subplots"""

    def custom_subplot_option(func):
        def callback(ctx, param, value):
            value = abs(value)
            nrows = value // 10
            nrows, ncols = divmod(nrows, 10)
            ctx.meta["n_col"] = ncols
            ctx.meta["n_row"] = nrows
            return value

        return click.option(
            "-sp",
            "--subplot",
            type=click.INT,
            default=dflt,
            show_default=True,
            help="The order of subplots.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_subplot_option


def cost_option(**kwargs):
    """Get option to get cost for FPR"""

    def custom_cost_option(func):
        def callback(ctx, param, value):
            if value < 0 or value > 1:
                raise click.BadParameter("Cost for FPR must be betwen 0 and 1")
            ctx.meta["cost"] = value
            return value

        return click.option(
            "-C",
            "--cost",
            type=float,
            default=0.99,
            show_default=True,
            help="Cost for FPR in minDCF",
            callback=callback,
            **kwargs,
        )(func)

    return custom_cost_option


def points_curve_option(**kwargs):
    """Get the number of points use to draw curves"""

    def custom_points_curve_option(func):
        def callback(ctx, param, value):
            if value < 2:
                raise click.BadParameter(
                    "Number of points to draw curves must be greater than 1",
                    ctx=ctx,
                )
            ctx.meta["points"] = value
            return value

        return click.option(
            "-n",
            "--points",
            type=INT,
            default=2000,
            show_default=True,
            help="The number of points use to draw curves in plots",
            callback=callback,
            **kwargs,
        )(func)

    return custom_points_curve_option


def n_bins_option(**kwargs):
    """Get the number of bins in the histograms"""
    possible_strings = [
        "auto",
        "fd",
        "doane",
        "scott",
        "rice",
        "sturges",
        "sqrt",
    ]

    def custom_n_bins_option(func):
        def callback(ctx, param, value):
            if value is None:
                value = ["doane"]
            else:
                tmp = value.split(",")
                try:
                    value = [
                        int(i) if i not in possible_strings else i for i in tmp
                    ]
                except Exception:
                    raise click.BadParameter("Incorrect number of bins inputs")
            ctx.meta["n_bins"] = value
            return value

        return click.option(
            "-b",
            "--nbins",
            type=click.STRING,
            default="doane",
            help="The number of bins for the different quantities to plot, "
            "seperated by commas. For example, if you plot histograms "
            "of negative and positive scores "
            ", input something like `100,doane`. All the "
            "possible bin options can be found in https://docs.scipy.org/doc/"
            "numpy/reference/generated/numpy.histogram.html. Be aware that "
            "for some corner cases, the option `auto` and `fd` can lead to "
            "MemoryError.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_n_bins_option


def table_option(dflt="rst", **kwargs):
    """Get table option for tabulate package
    More informnations: https://pypi.org/project/tabulate/
    """

    def custom_table_option(func):
        def callback(ctx, param, value):
            ctx.meta["tablefmt"] = value
            return value

        return click.option(
            "--tablefmt",
            type=click.Choice(tabulate.tabulate_formats),
            default=dflt,
            show_default=True,
            help="Format of printed tables.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_table_option


def output_plot_file_option(default_out="plots.pdf", **kwargs):
    """Get options for output file for plots"""

    def custom_output_plot_file_option(func):
        def callback(ctx, param, value):
            """Save ouput file  and associated pdf in context list,
            print the path of the file in the log"""
            ctx.meta["output"] = value
            ctx.meta["PdfPages"] = PdfPages(value)
            LOGGER.debug("Plots will be output in %s", value)
            return value

        return click.option(
            "-o",
            "--output",
            default=default_out,
            show_default=True,
            help="The file to save the plots in.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_output_plot_file_option


def output_log_metric_option(**kwargs):
    """Get options for output file for metrics"""

    def custom_output_log_file_option(func):
        def callback(ctx, param, value):
            if value is not None:
                LOGGER.debug("Metrics will be output in %s", value)
            ctx.meta["log"] = value
            return value

        return click.option(
            "-l",
            "--log",
            default=None,
            type=click.STRING,
            help="If provided, computed numbers are written to "
            "this file instead of the standard output.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_output_log_file_option


def no_line_option(**kwargs):
    """Get option flag to say if no line should be displayed"""

    def custom_no_line_option(func):
        def callback(ctx, param, value):
            ctx.meta["no_line"] = value
            return value

        return click.option(
            "--no-line",
            is_flag=True,
            default=False,
            show_default=True,
            help="If set does not display vertical lines",
            callback=callback,
            **kwargs,
        )(func)

    return custom_no_line_option


def criterion_option(
    lcriteria=["eer", "min-hter", "far"], check=True, **kwargs
):
    """Get option flag to tell which criteriom is used (default:eer)

    Parameters
    ----------
    lcriteria : :any:`list`
        List of possible criteria
    """

    def custom_criterion_option(func):
        list_accepted_crit = (
            lcriteria if lcriteria is not None else ["eer", "min-hter", "far"]
        )

        def callback(ctx, param, value):
            if value not in list_accepted_crit and check:
                raise click.BadParameter(
                    "Incorrect value for `--criterion`. "
                    "Must be one of [`%s`]" % "`, `".join(list_accepted_crit)
                )
            ctx.meta["criterion"] = value
            return value

        return click.option(
            "-c",
            "--criterion",
            default="eer",
            help="Criterion to compute plots and "
            "metrics: %s)" % ", ".join(list_accepted_crit),
            callback=callback,
            is_eager=True,
            show_default=True,
            **kwargs,
        )(func)

    return custom_criterion_option


def decimal_option(dflt=1, short="-d", **kwargs):
    """Get option to get decimal value"""

    def custom_decimal_option(func):
        def callback(ctx, param, value):
            ctx.meta["decimal"] = value
            return value

        return click.option(
            short,
            "--decimal",
            type=click.INT,
            default=dflt,
            help="Number of decimals to be printed.",
            callback=callback,
            show_default=True,
            **kwargs,
        )(func)

    return custom_decimal_option


def far_option(far_name="FAR", **kwargs):
    """Get option to get far value"""

    def custom_far_option(func):
        def callback(ctx, param, value):
            if value is not None and (value > 1 or value < 0):
                raise click.BadParameter(
                    "{} value should be between 0 and 1".format(far_name)
                )
            ctx.meta["far_value"] = value
            return value

        return click.option(
            "-f",
            "--{}-value".format(far_name.lower()),
            "far_value",
            type=click.FLOAT,
            default=None,
            help="The {} value for which to compute threshold. This option "
            "must be used alongside `--criterion far`.".format(far_name),
            callback=callback,
            show_default=True,
            **kwargs,
        )(func)

    return custom_far_option


def min_far_option(far_name="FAR", dflt=1e-4, **kwargs):
    """Get option to get min far value"""

    def custom_min_far_option(func):
        def callback(ctx, param, value):
            if value is not None and (value > 1 or value < 0):
                raise click.BadParameter(
                    "{} value should be between 0 and 1".format(far_name)
                )
            ctx.meta["min_far_value"] = value
            return value

        return click.option(
            "-M",
            "--min-{}-value".format(far_name.lower()),
            "min_far_value",
            type=click.FLOAT,
            default=dflt,
            help="Select the minimum {} value used in ROC and DET plots; "
            "should be a power of 10.".format(far_name),
            callback=callback,
            show_default=True,
            **kwargs,
        )(func)

    return custom_min_far_option


def figsize_option(dflt="4,3", **kwargs):
    """Get option for matplotlib figsize

    Parameters
    ----------
    dflt : str
        matplotlib default figsize for the command. must be a a list of int
        separated by commas.

    Returns
    -------
    callable
        A decorator to be used for adding score arguments for click commands
    """

    def custom_figsize_option(func):
        def callback(ctx, param, value):
            ctx.meta["figsize"] = (
                value if value is None else [float(x) for x in value.split(",")]
            )
            if value is not None:
                plt.rcParams["figure.figsize"] = ctx.meta["figsize"]
            return value

        return click.option(
            "--figsize",
            default=dflt,
            show_default=True,
            help="If given, will run "
            "``plt.rcParams['figure.figsize']=figsize)``. "
            "Example: --figsize 4,6",
            callback=callback,
            **kwargs,
        )(func)

    return custom_figsize_option


def legend_loc_option(dflt="best", **kwargs):
    """Get the legend location of the plot"""

    def custom_legend_loc_option(func):
        def callback(ctx, param, value):
            ctx.meta["legend_loc"] = value.replace("-", " ") if value else value
            return value

        return click.option(
            "-ll",
            "--legend-loc",
            default=dflt,
            show_default=True,
            type=click.Choice(
                [
                    "best",
                    "upper-right",
                    "upper-left",
                    "lower-left",
                    "lower-right",
                    "right",
                    "center-left",
                    "center-right",
                    "lower-center",
                    "upper-center",
                    "center",
                ]
            ),
            help="The legend location code",
            callback=callback,
            **kwargs,
        )(func)

    return custom_legend_loc_option


def line_width_option(**kwargs):
    """Get line width option for the plots"""

    def custom_line_width_option(func):
        def callback(ctx, param, value):
            ctx.meta["line_width"] = value
            return value

        return click.option(
            "--line-width",
            type=FLOAT,
            help="The line width of plots",
            callback=callback,
            **kwargs,
        )(func)

    return custom_line_width_option


def marker_style_option(**kwargs):
    """Get marker style otpion for the plots"""

    def custom_marker_style_option(func):
        def callback(ctx, param, value):
            ctx.meta["marker_style"] = value
            return value

        return click.option(
            "--marker-style",
            type=FLOAT,
            help="The marker style of the plots",
            callback=callback,
            **kwargs,
        )(func)

    return custom_marker_style_option


def legends_option(**kwargs):
    """Get the legends option for the different systems"""

    def custom_legends_option(func):
        def callback(ctx, param, value):
            if value is not None:
                value = value.split(",")
            ctx.meta["legends"] = value
            return value

        return click.option(
            "-lg",
            "--legends",
            type=click.STRING,
            default=None,
            help="The legend for each system comma separated. "
            "Example: --legends ISV,CNN",
            callback=callback,
            **kwargs,
        )(func)

    return custom_legends_option


def title_option(**kwargs):
    """Get the title option for the different systems"""

    def custom_title_option(func):
        def callback(ctx, param, value):
            ctx.meta["title"] = value
            return value

        return click.option(
            "-t",
            "--title",
            type=click.STRING,
            default=None,
            help="The title of the plots. Provide just a space (-t ' ') to "
            "remove the titles from figures.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_title_option


def titles_option(**kwargs):
    """Get the titles option for the different plots"""

    def custom_title_option(func):
        def callback(ctx, param, value):
            if value is not None:
                value = value.split(",")
            ctx.meta["titles"] = value or []
            return value or []

        return click.option(
            "-ts",
            "--titles",
            type=click.STRING,
            default=None,
            help="The titles of the plots seperated by commas. "
            'For example, if the figure has two plots, "MyTitleA,MyTitleB" '
            "is a possible input."
            " Provide just a space (-ts ' ') to "
            "remove the titles from figures.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_title_option


def x_label_option(dflt=None, **kwargs):
    """Get the label option for X axis"""

    def custom_x_label_option(func):
        def callback(ctx, param, value):
            ctx.meta["x_label"] = value
            return value

        return click.option(
            "-xl",
            "--x-label",
            type=click.STRING,
            default=dflt,
            show_default=True,
            help="Label for x-axis",
            callback=callback,
            **kwargs,
        )(func)

    return custom_x_label_option


def y_label_option(dflt=None, **kwargs):
    """Get the label option for Y axis"""

    def custom_y_label_option(func):
        def callback(ctx, param, value):
            ctx.meta["y_label"] = value
            return value

        return click.option(
            "-yl",
            "--y-label",
            type=click.STRING,
            default=dflt,
            help="Label for y-axis",
            callback=callback,
            **kwargs,
        )(func)

    return custom_y_label_option


def style_option(**kwargs):
    """Get option for matplotlib style"""

    def custom_style_option(func):
        def callback(ctx, param, value):
            ctx.meta["style"] = value
            plt.style.use(value)
            return value

        return click.option(
            "--style",
            multiple=True,
            type=click.types.Choice(sorted(plt.style.available)),
            help="The matplotlib style to use for plotting. You can provide "
            "multiple styles by repeating this option",
            callback=callback,
            **kwargs,
        )(func)

    return custom_style_option


def metrics_command(
    docstring,
    criteria=("eer", "min-hter", "far"),
    far_name="FAR",
    check_criteria=True,
    **kwarg,
):
    def custom_metrics_command(func):
        func.__doc__ = docstring

        @click.command(**kwarg)
        @scores_argument(nargs=-1)
        @eval_option()
        @table_option()
        @output_log_metric_option()
        @criterion_option(criteria, check=check_criteria)
        @thresholds_option()
        @far_option(far_name=far_name)
        @legends_option()
        @open_file_mode_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @decimal_option()
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_metrics_command


METRICS_HELP = """Prints a table that contains {names} for a given
    threshold criterion ({criteria}).
    {hter_note}

    You need to provide one or more development score file(s) for each
    experiment. You can also provide evaluation files along with dev files. If
    evaluation scores are provided, you must use flag `--eval`.

    {score_format}

    Resulting table format can be changed using the `--tablefmt`.

    Examples:

        $ {command} -v scores-dev

        $ {command} -e -l results.txt sys1/scores-{{dev,eval}}

        $ {command} -e {{sys1,sys2}}/scores-{{dev,eval}}

    """


def roc_command(docstring, far_name="FAR"):
    def custom_roc_command(func):
        func.__doc__ = docstring

        @click.command()
        @scores_argument(nargs=-1)
        @titles_option()
        @legends_option()
        @no_legend_option()
        @legend_loc_option(dflt=None)
        @sep_dev_eval_option()
        @output_plot_file_option(default_out="roc.pdf")
        @eval_option()
        @tpr_option(True)
        @semilogx_option(True)
        @lines_at_option()
        @axes_val_option()
        @min_far_option(far_name=far_name)
        @x_rotation_option()
        @x_label_option()
        @y_label_option()
        @points_curve_option()
        @const_layout_option()
        @figsize_option()
        @style_option()
        @linestyles_option()
        @alpha_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_roc_command


ROC_HELP = """Plot ROC (receiver operating characteristic) curve.
    The plot will represent the false match rate on the horizontal axis and the
    false non match rate on the vertical axis.  The values for the axis will be
    computed using :py:func:`bob.measure.roc`.

    You need to provide one or more development score file(s) for each
    experiment. You can also provide evaluation files along with dev files. If
    evaluation scores are provided, you must use flag `--eval`.

    {score_format}

    Examples:

        $ {command} -v scores-dev

        $ {command} -e -v sys1/scores-{{dev,eval}}

        $ {command} -e -v -o my_roc.pdf {{sys1,sys2}}/scores-{{dev,eval}}
    """


def det_command(docstring, far_name="FAR"):
    def custom_det_command(func):
        func.__doc__ = docstring

        @click.command()
        @scores_argument(nargs=-1)
        @output_plot_file_option(default_out="det.pdf")
        @titles_option()
        @legends_option()
        @no_legend_option()
        @legend_loc_option(dflt="upper-right")
        @sep_dev_eval_option()
        @eval_option()
        @axes_val_option(dflt="0.01,95,0.01,95")
        @min_far_option(far_name=far_name)
        @x_rotation_option(dflt=45)
        @x_label_option()
        @y_label_option()
        @points_curve_option()
        @lines_at_option()
        @const_layout_option()
        @figsize_option()
        @style_option()
        @linestyles_option()
        @alpha_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_det_command


DET_HELP = """Plot DET (detection error trade-off) curve.
    modified ROC curve which plots error rates on both axes
    (false positives on the x-axis and false negatives on the y-axis).

    You need to provide one or more development score file(s) for each
    experiment. You can also provide evaluation files along with dev files. If
    evaluation scores are provided, you must use flag `--eval`.

    {score_format}

    Examples:

        $ {command} -v scores-dev

        $ {command} -e -v sys1/scores-{{dev,eval}}

        $ {command} -e -v -o my_det.pdf {{sys1,sys2}}/scores-{{dev,eval}}
    """


def epc_command(docstring):
    def custom_epc_command(func):
        func.__doc__ = docstring

        @click.command()
        @scores_argument(min_arg=1, force_eval=True, nargs=-1)
        @output_plot_file_option(default_out="epc.pdf")
        @titles_option()
        @legends_option()
        @no_legend_option()
        @legend_loc_option(dflt="upper-center")
        @points_curve_option()
        @const_layout_option()
        @x_label_option()
        @y_label_option()
        @figsize_option()
        @style_option()
        @linestyles_option()
        @alpha_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_epc_command


EPC_HELP = """Plot EPC (expected performance curve).
    plots the error rate on the eval set depending on a threshold selected
    a-priori on the development set and accounts for varying relative cost
    in [0; 1] of FPR and FNR when calculating the threshold.

    You need to provide one or more development score and eval file(s)
    for each experiment.

    {score_format}

    Examples:

        $ {command} -v scores-{{dev,eval}}

        $ {command} -v -o my_epc.pdf {{sys1,sys2}}/scores-{{dev,eval}}
    """


def hist_command(docstring, far_name="FAR"):
    def custom_hist_command(func):
        func.__doc__ = docstring

        @click.command()
        @scores_argument(nargs=-1)
        @output_plot_file_option(default_out="hist.pdf")
        @eval_option()
        @hide_dev_option()
        @n_bins_option()
        @titles_option()
        @no_legend_option()
        @legend_ncols_option()
        @criterion_option()
        @far_option(far_name=far_name)
        @no_line_option()
        @thresholds_option()
        @subplot_option()
        @const_layout_option()
        @print_filenames_option()
        @figsize_option(dflt=None)
        @style_option()
        @x_label_option()
        @y_label_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_hist_command


HIST_HELP = """ Plots histograms of positive and negatives along with threshold
    criterion.

    You need to provide one or more development score file(s) for each
    experiment. You can also provide evaluation files along with dev files. If
    evaluation scores are provided, you must use the `--eval` flag. The
    threshold is always computed from development score files.

    By default, when eval-scores are given, only eval-scores histograms are
    displayed with threshold line computed from dev-scores.

    {score_format}

    Examples:

        $ {command} -v scores-dev

        $ {command} -e -v sys1/scores-{{dev,eval}}

        $ {command} -e -v --criterion min-hter {{sys1,sys2}}/scores-{{dev,eval}}
    """


def evaluate_command(
    docstring, criteria=("eer", "min-hter", "far"), far_name="FAR"
):
    def custom_evaluate_command(func):
        func.__doc__ = docstring

        @click.command()
        @scores_argument(nargs=-1)
        @legends_option()
        @sep_dev_eval_option()
        @table_option()
        @eval_option()
        @criterion_option(criteria)
        @far_option(far_name=far_name)
        @output_log_metric_option()
        @output_plot_file_option(default_out="eval_plots.pdf")
        @lines_at_option()
        @points_curve_option()
        @const_layout_option()
        @figsize_option(dflt=None)
        @style_option()
        @linestyles_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_evaluate_command


EVALUATE_HELP = """Runs error analysis on score sets.

    \b
    1. Computes the threshold using a criteria (EER by default) on
       development set scores
    2. Applies the above threshold on evaluation set scores to compute the
       HTER if a eval-score (use --eval) set is provided.
    3. Reports error rates on the console or in a log file.
    4. Plots ROC, DET, and EPC curves and score distributions to a multi-page
       PDF file

    You need to provide 1 or 2 score files for each biometric system in this
    order:

    \b
    * development scores
    * evaluation scores

    {score_format}

    Examples:

        $ {command} -v dev-scores

        $ {command} -v /path/to/sys-{{1,2,3}}/scores-dev

        $ {command} -e -v /path/to/sys-{{1,2,3}}/scores-{{dev,eval}}

        $ {command} -v -l metrics.txt -o my_plots.pdf dev-scores

    This command is a combination of metrics, roc, det, epc, and hist commands.
    If you want more flexibility in your plots, please use the individual
    commands.
    """


def evaluate_flow(
    ctx, scores, evaluation, metrics, roc, det, epc, hist, **kwargs
):
    # open_mode is always write in this command.
    ctx.meta["open_mode"] = "w"
    criterion = ctx.meta.get("criterion")
    if criterion is not None:
        click.echo("Computing metrics with %s..." % criterion)
        ctx.invoke(metrics, scores=scores, evaluation=evaluation)
        if ctx.meta.get("log") is not None:
            click.echo("[metrics] => %s" % ctx.meta["log"])

    # avoid closing pdf file before all figures are plotted
    ctx.meta["closef"] = False
    if evaluation:
        click.echo("Starting evaluate with dev and eval scores...")
    else:
        click.echo("Starting evaluate with dev scores only...")
    click.echo("Computing ROC...")
    # set axes limits for ROC
    ctx.forward(roc)  # use class defaults plot settings
    click.echo("Computing DET...")
    ctx.forward(det)  # use class defaults plot settings
    if evaluation:
        click.echo("Computing EPC...")
        ctx.forward(epc)  # use class defaults plot settings
    # the last one closes the file
    ctx.meta["closef"] = True
    click.echo("Computing score histograms...")
    ctx.meta["criterion"] = "eer"  # no criterion passed in evaluate
    ctx.forward(hist)
    click.echo("Evaluate successfully completed!")
    click.echo("[plots] => %s" % (ctx.meta["output"]))


def n_protocols_option(required=True, **kwargs):
    """Get option for number of protocols."""

    def custom_n_protocols_option(func):
        def callback(ctx, param, value):
            value = abs(value)
            ctx.meta["protocols_number"] = value
            return value

        return click.option(
            "-pn",
            "--protocols-number",
            type=click.INT,
            show_default=True,
            required=required,
            help="The number of protocols of cross validation.",
            callback=callback,
            **kwargs,
        )(func)

    return custom_n_protocols_option


def multi_metrics_command(
    docstring, criteria=("eer", "min-hter", "far"), far_name="FAR", **kwargs
):
    def custom_metrics_command(func):
        func.__doc__ = docstring

        @click.command("multi-metrics", **kwargs)
        @scores_argument(nargs=-1)
        @eval_option()
        @n_protocols_option()
        @table_option()
        @output_log_metric_option()
        @criterion_option(criteria)
        @thresholds_option()
        @far_option(far_name=far_name)
        @legends_option()
        @open_file_mode_option()
        @verbosity_option(LOGGER)
        @click.pass_context
        @decimal_option()
        @functools.wraps(func)
        def wrapper(*args, **kwds):
            return func(*args, **kwds)

        return wrapper

    return custom_metrics_command


MULTI_METRICS_HELP = """Multi protocol (cross-validation) metrics.

    Prints a table that contains mean and standard deviation of {names} for a given
    threshold criterion ({criteria}). The metrics are averaged over several protocols.
    The idea is that each protocol corresponds to one fold in your cross-validation.

    You need to provide as many as development score files as the number of
    protocols per system. You can also provide evaluation files along with dev
    files. If evaluation scores are provided, you must use flag `--eval`. The
    number of protocols must be provided using the `--protocols-number` option.

    {score_format}

    Resulting table format can be changed using the `--tablefmt`.

    Examples:

        $ {command} -vv -pn 3 {{p1,p2,p3}}/scores-dev

        $ {command} -vv -pn 3 -e {{p1,p2,p3}}/scores-{{dev,eval}}

        $ {command} -vv -pn 3 -e {{sys1,sys2}}/{{p1,p2,p3}}/scores-{{dev,eval}}
    """



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

import numpy

from bob.measure.utils import confidence_for_indicator_variable


def test_confidence_interval():
    def assert_confidence(x, n, expected_lower, expected_upper):
        lower, upper = confidence_for_indicator_variable(x, n)
        assert numpy.allclose(lower, expected_lower)
        assert numpy.allclose(upper, expected_upper)

    assert_confidence(1, 2, 0.01257911709342505, 0.98742088290657493)
    assert_confidence(10, 10, 0.69150289218123917, 1)
    assert_confidence(0, 10, 0, 0.30849710781876077)



""" utility functions for bob.measure """

import logging

import numpy
import scipy.stats

LOGGER = logging.getLogger(__name__)


def remove_nan(scores):
    """remove_nan

    Remove NaN(s) in the given array

    Parameters
    ----------
    scores :
        :py:class:`numpy.ndarray` : array

    Returns
    -------
        :py:class:`numpy.ndarray` : array without NaN(s)
        :py:class:`int` : number of NaN(s) in the input array
        :py:class:`int` : length of the input array
    """
    nans = numpy.isnan(scores)
    sum_nans = sum(nans)
    total = len(scores)
    if sum_nans > 0:
        LOGGER.warning("Found {} NaNs in {} scores".format(sum_nans, total))
    return scores[~nans], sum_nans, total


def get_fta(scores):
    """get_fta
        calculates the Failure To Acquire (FtA) rate, i.e. proportion of NaN(s)
        in the input scores

    Parameters
    ----------
    scores :
        Tuple of (``positive``, ``negative``) :py:class:`numpy.ndarray`.

    Returns
    -------
    (:py:class:`numpy.ndarray`, :py:class:`numpy.ndarray`): scores without
    NaN(s)
    :py:class:`float` : failure to acquire rate
    """
    fta_sum, fta_total = 0.0, 0.0
    neg, sum_nans, total = remove_nan(scores[0])
    fta_sum += sum_nans
    fta_total += total
    pos, sum_nans, total = remove_nan(scores[1])
    fta_sum += sum_nans
    fta_total += total
    return ((neg, pos), fta_sum / fta_total)


def get_fta_list(scores):
    """Get FTAs for a list of scores

    Parameters
    ----------
    scores: :any:`list`
        list of scores

    Returns
    -------
    neg_list: :any:`list`
        list of negatives
    pos_list: :any:`list`
        list of positives
    fta_list: :any:`list`
        list of FTAs
    """
    neg_list = []
    pos_list = []
    fta_list = []
    for score in scores:
        neg = pos = fta = None
        if score is not None:
            (neg, pos), fta = get_fta(score)
            if neg is None:
                raise ValueError("While loading dev-score file")
        neg_list.append(neg)
        pos_list.append(pos)
        fta_list.append(fta)
    return (neg_list, pos_list, fta_list)


def get_thres(criter, neg, pos, far=None):
    """Get threshold for the given positive/negatives scores and criterion

    Parameters
    ----------
    criter :
        Criterion (`eer` or `hter` or `far`)
    neg : :py:class:`numpy.ndarray`:
        array of negative scores
        pos : :py:class:`numpy.ndarray`::
        array of positive scores

    Returns
    -------
    :py:obj:`float`
        threshold
    """
    if criter == "eer":
        from . import eer_threshold

        return eer_threshold(neg, pos)
    elif criter == "min-hter":
        from . import min_hter_threshold

        return min_hter_threshold(neg, pos)
    elif criter == "far":
        if far is None:
            raise ValueError(
                "FAR value must be provided through "
                "``--far-value`` or ``--fpr-value`` option."
            )
        from . import far_threshold

        return far_threshold(neg, pos, far)
    else:
        raise ValueError("Incorrect plotting criterion: ``%s``" % criter)


def get_colors(n):
    """get_colors
    Get a list of matplotlib colors

    Parameters
    ----------
    n : :obj:`int`
        Number of colors to output

    Returns
    -------
    :any:`list`
        list of colors
    """
    if n > 10:
        from matplotlib import pyplot

        cmap = pyplot.cm.get_cmap(name="magma")
        return [cmap(i) for i in numpy.linspace(0, 1.0, n + 1)]

    return ["C0", "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9"]


def get_linestyles(n, on=True):
    """Get a list of matplotlib linestyles

    Parameters
    ----------
    n : :obj:`int`
        Number of linestyles to output

    Returns
    -------
    :any:`list`
        list of linestyles
    """
    if not on:
        return [None] * n

    list_linestyles = [
        (0, ()),  # solid
        (0, (1, 1)),  # densely dotted
        (0, (5, 5)),  # dashed
        (0, (5, 1)),  # densely dashed
        (0, (3, 1, 1, 1, 1, 1)),  # densely dashdotdotted
        (0, (3, 10, 1, 10, 1, 10)),  # loosely dashdotdotted
        (0, (3, 5, 1, 5, 1, 5)),  # dashdotdotted
        (0, (3, 1, 1, 1)),  # densely dashdotted
        (0, (1, 5)),  # dotted
        (0, (3, 5, 1, 5)),  # dashdotted
        (0, (5, 10)),  # loosely dashed
        (0, (3, 10, 1, 10)),  # loosely dashdotted
        (0, (1, 10)),  # loosely dotted
    ]
    while n > len(list_linestyles):
        list_linestyles += list_linestyles
    return list_linestyles


def confidence_for_indicator_variable(x, n, alpha=0.05):
    """Calculates the confidence interval for proportion estimates
    The Clopper-Pearson interval method is used for estimating the confidence
    intervals.

    Parameters
    ----------
    x : int
        The number of successes.
    n : int
        The number of trials.
        alpha : :obj:`float`, optional
        The 1-confidence value that you want. For example, alpha should be 0.05
        to obtain 95% confidence intervals.

    Returns
    -------
    (:obj:`float`, :obj:`float`)
        a tuple of (lower_bound, upper_bound) which
        shows the limit of your success rate: lower_bound < x/n < upper_bound
    """
    lower_bound = scipy.stats.beta.ppf(alpha / 2.0, x, n - x + 1)
    upper_bound = scipy.stats.beta.ppf(1 - alpha / 2.0, x + 1, n - x)
    if numpy.isnan(lower_bound):
        lower_bound = 0
    if numpy.isnan(upper_bound):
        upper_bound = 1
    return (lower_bound, upper_bound)



# import Libraries of other lib packages
import numpy

from . import calibration, load, plot  # noqa: F401
from ._library import *  # noqa: F401, F403
from ._library import eer_threshold, farfrr, logger, roc


def fprfnr(negatives, positives, threshold):
    """Alias for :py:func:`bob.measure.farfrr`"""
    return farfrr(negatives, positives, threshold)


def mse(estimation, target):
    r"""Mean square error between a set of outputs and target values

    Uses the formula:

    .. math::

      MSE(\hat{\Theta}) = E[(\hat{\Theta} - \Theta)^2]

    Estimation (:math:`\hat{\Theta}`) and target (:math:`\Theta`) are supposed to
    have 2 dimensions. Different examples are organized as rows while different
    features in the estimated values or targets are organized as different
    columns.


    Parameters:

      estimation (array): an N-dimensional array that corresponds to the value
        estimated by your procedure

      target (array): an N-dimensional array that corresponds to the expected
        value


    Returns:

      float: The average of the squared error between the estimated value and the
      target

    """
    return numpy.mean((estimation - target) ** 2, 0)


def rmse(estimation, target):
    r"""Calculates the root mean square error between a set of outputs and target

    Uses the formula:

    .. math::

      RMSE(\hat{\Theta}) = \sqrt(E[(\hat{\Theta} - \Theta)^2])

    Estimation (:math:`\hat{\Theta}`) and target (:math:`\Theta`) are supposed to
    have 2 dimensions. Different examples are organized as rows while different
    features in the estimated values or targets are organized as different
    columns.


    Parameters:

      estimation (array): an N-dimensional array that corresponds to the value
        estimated by your procedure

      target (array): an N-dimensional array that corresponds to the expected
        value


    Returns:

      float: The square-root of the average of the squared error between the
      estimated value and the target

    """
    return numpy.sqrt(mse(estimation, target))


def relevance(input, machine):
    """Calculates the relevance of every input feature to the estimation process

    Uses the formula:

      Neural Triggering System Operating on High Resolution Calorimetry
      Information, Anjos et al, April 2006, Nuclear Instruments and Methods in
      Physics Research, volume 559, pages 134-138

    .. math::

      R(x_{i}) = |E[(o(x) - o(x|x_{i}=E[x_{i}]))^2]|

    In other words, the relevance of a certain input feature **i** is the change
    on the machine output value when such feature is replaced by its mean for all
    input vectors. For this to work, the `input` parameter has to be a 2D array
    with features arranged column-wise while different examples are arranged
    row-wise.


    Parameters:

      input (array): an N-dimensional array that corresponds to the value
        estimated by your model

      machine (object): A machine that can be called to "process" your input


    Returns:

      array: An 1D float array as large as the number of columns (second
      dimension) of your input array, estimating the "relevance" of each input
      column (or feature) to the score provided by the machine.

    """

    o = machine(input)
    i2 = input.copy()
    retval = numpy.ndarray((input.shape[1],), "float64")
    retval.fill(0)
    for k in range(input.shape[1]):
        i2[:, :] = input  # reset
        i2[:, k] = numpy.mean(input[:, k])
        retval[k] = (mse(machine(i2), o).sum()) ** 0.5

    return retval


def recognition_rate(cmc_scores, threshold=None, rank=1):
    """Calculates the recognition rate from the given input

    It is identical to the CMC value for the given ``rank``.

    The input has a specific format, which is a list of two-element tuples.  Each
    of the tuples contains the negative :math:`\\{S_p^-\\}` and the positive
    :math:`\\{S_p^+\\}` scores for one probe item :math:`p`, or ``None`` in case
    of open set recognition.

    If ``threshold`` is set to ``None``, the rank 1 recognition rate is defined
    as the number of test items, for which the highest positive
    :math:`\\max\\{S_p^+\\}` score is greater than or equal to all negative
    scores, divided by the number of all probe items :math:`P`:

    .. math::

      \\mathrm{RR} = \\frac{1}{P} \\sum_{p=1}^{P} \\begin{cases} 1 & \\mathrm{if } \\max\\{S_p^+\\} >= \\max\\{S_p^-\\}\\\\ 0 & \\mathrm{otherwise} \\end{cases}

    For a given rank :math:`r>1`, up to :math:`r` negative scores that are higher
    than the highest positive score are allowed to still count as correctly
    classified in the top :math:`r` rank.

    If ``threshold`` :math:`\\theta` is given, **all** scores below threshold
    will be filtered out.  Hence, if all positive scores are below threshold
    :math:`\\max\\{S_p^+\\} < \\theta`, the probe will be misclassified **at any
    rank**.

    For open set recognition, i.e., when there exist a tuple including negative
    scores without corresponding positive scores (``None``), and **all** negative
    scores are below ``threshold`` :math:`\\max\\{S_p^+\\} < \\theta`, the probe
    item is correctly rejected, **and it does not count into the denominator**
    :math:`P`.  When no ``threshold`` is provided, the open set probes will
    **always** count as misclassified, regardless of the ``rank``.

    .. warn:
       For open set tests, this rate does not correspond to a standard rate.
       Please use :py:func:`detection_identification_rate` and
       :py:func:`false_alarm_rate` instead.


    Parameters:

      cmc_scores (:py:class:`list`): A list in the format ``[(negatives,
        positives), ...]`` containing the CMC scores (i.e. :py:class:`list`:
        A list of tuples, where each tuple contains the
        ``negative`` and ``positive`` scores for one probe of the database).

        Each pair contains the ``negative`` and the ``positive`` scores for **one
        probe item**.  Each pair can contain up to one empty array (or ``None``),
        i.e., in case of open set recognition.

      threshold (:obj:`float`, optional): Decision threshold. If not ``None``, **all**
        scores will be filtered by the threshold. In an open set recognition
        problem, all open set scores (negatives with no corresponding positive)
        for which all scores are below threshold, will be counted as correctly
        rejected and **removed** from the probe list (i.e., the denominator).

      rank (:obj:`int`, optional):
        The rank for which the recognition rate should be computed, 1 by default.


    Returns:

      float: The (open set) recognition rate for the given rank, a value between
      0 and 1.

    """
    # If no scores are given, the recognition rate is exactly 0.
    if not cmc_scores:
        return 0.0

    correct = 0
    counter = 0
    for neg, pos in cmc_scores:
        # set all values that are empty before to None
        if pos is not None and not numpy.array(pos).size:
            pos = None
        if neg is not None and not numpy.array(neg).size:
            neg = None

        if pos is None and neg is None:
            raise ValueError(
                "One pair of the CMC scores has neither positive nor negative values"
            )

        # filter out any negative or positive scores below threshold; scores with exactly the threshold are also filtered out
        # now, None and an empty array have different meanings.
        if threshold is not None:
            if neg is not None:
                neg = numpy.array(neg)[neg > threshold]
            if pos is not None:
                pos = numpy.array(pos)[pos > threshold]

        if pos is None:
            # no positives, so we definitely do not have a match;
            # check if we have negatives above threshold
            if not neg.size:
                # we have no negative scores over the threshold, so we have correctly rejected the probe
                # don't increase any of the two counters...
                continue
            # we have negatives over threshold, so we have incorrect classifications; independent on the actual rank
            counter += 1
        else:
            # we have a positive, so we need to count the probe
            counter += 1

            if not numpy.array(pos).size:
                # all positive scores have been filtered out by the threshold, we definitely have a mis-match
                continue

            # get the maximum positive score for the current probe item
            # (usually, there is only one positive score, but just in case...)
            max_pos = numpy.max(pos)

            if neg is None or not numpy.array(neg).size:
                # if we had no negatives, or all negatives were below threshold, we have a match at rank 1
                correct += 1
            else:
                # count the number of negative scores that are higher than the best positive score
                index = numpy.sum(neg >= max_pos)
                if index < rank:
                    correct += 1

    return float(correct) / float(counter)


def cmc(cmc_scores):
    """Calculates the cumulative match characteristic (CMC) from the given input.

    The input has a specific format, which is a list of two-element tuples. Each
    of the tuples contains the negative and the positive scores for one probe
    item.

    For each probe item the probability that the rank :math:`r` of the positive
    score is calculated.  The rank is computed as the number of negative scores
    that are higher than the positive score.  If several positive scores for one
    test item exist, the **highest** positive score is taken. The CMC finally
    computes how many test items have rank r or higher, divided by the total
    number of test values.

    .. note::

       The CMC is not available for open set classification. Please use the
       :py:func:`detection_identification_rate` and :py:func:`false_alarm_rate`
       instead.


    Parameters
    ----------

    cmc_scores : :py:class:`list`
      A list in the format ``[(negatives, positives), ...]`` containing the CMC
      scores.

      Each pair contains the ``negative`` and the ``positive`` scores for **one
      probe item**.  Each pair can contain up to one empty array (or ``None``),
      i.e., in case of open set recognition.


    Returns
    -------

    1D :py:class:`numpy.ndarray` of `float`
      A 1D float array representing the CMC curve.
      The rank 1 recognition rate can be found in ``array[0]``, rank 2 rate in
      ``array[1]``, and so on. The number of ranks (``array.shape[0]``) is the
      number of gallery items. Values are in range ``[0,1]``.
    """

    # If no scores are given, we cannot plot anything
    probe_count = float(len(cmc_scores))
    if not probe_count:
        raise ValueError("The given set of scores is empty")

    # compute MC
    match_characteristic = numpy.zeros(
        (max([len(neg) for neg, _ in cmc_scores if neg is not None]) + 1,),
        numpy.int,
    )

    for neg, pos in cmc_scores:
        if pos is None or not numpy.array(pos).size:
            raise ValueError(
                "For the CMC computation at least one positive score per pair is necessary."
            )
        if neg is None:
            neg = []

        # get the maximum positive score for the current probe item
        # (usually, there is only one positive score, but just in case...)
        max_pos = numpy.max(pos)

        # count the number of negative scores that are higher than the best positive score
        index = numpy.sum(neg >= max_pos)
        match_characteristic[index] += 1

    # cumulate
    cumulative_match_characteristic = numpy.cumsum(
        match_characteristic, dtype=numpy.float64
    )
    return cumulative_match_characteristic / probe_count


def detection_identification_rate(cmc_scores, threshold, rank=1):
    """Computes the `detection and identification rate` for the given threshold.

    This value is designed to be used in an open set identification protocol, and
    defined in Chapter 14.1 of [LiJain2005]_.

    Although the detection and identification rate is designed to be computed on
    an open set protocol, it uses only the probe elements, for which a
    corresponding gallery element exists.  For closed set identification
    protocols, this function is identical to :py:func:`recognition_rate`.  The
    only difference is that for this function, a ``threshold`` for the scores
    need to be defined, while for :py:func:`recognition_rate` it is optional.


    Parameters:

      cmc_scores (:py:class:`list`): A list in the format ``[(negatives,
        positives), ...]`` containing the CMC.

        Each pair contains the ``negative`` and the ``positive`` scores for **one
        probe item**.  Each pair can contain up to one empty array (or ``None``),
        i.e., in case of open set recognition.

      threshold (float): The decision threshold :math:`\\tau``.

      rank (:obj:`int`, optional): The rank for which the curve should be plotted


    Returns:

      float: The detection and identification rate for the given threshold.

    """

    # count the correctly classifier probes
    correct = 0
    counter = 0
    for neg, pos in cmc_scores:
        if pos is None or not numpy.array(pos).size:
            # we only consider probes with corresponding gallery items
            continue
        # we have an in-gallery probe
        counter += 1
        # check, if it is correctly classified
        if neg is None:
            neg = []

        # get the maximum positive score for the current probe item
        # (usually, there is only one positive score, but just in case...)
        max_pos = numpy.max(pos)

        index = numpy.sum(
            neg >= max_pos
        )  # compute the rank (in fact, rank - 1)
        if max_pos >= threshold and index < rank:
            correct += 1

    if not counter:
        logger.warn("No in-gallery probe was found")
        return 0.0

    return float(correct) / float(counter)


def false_alarm_rate(cmc_scores, threshold):
    """Computes the `false alarm rate` for the given threshold,.

    This value is designed to be used in an open set identification protocol, and
    defined in Chapter 14.1 of [LiJain2005]_.

    The false alarm rate is designed to be computed on an open set protocol, it
    uses only the probe elements, for which **no** corresponding gallery element
    exists.


    Parameters:

      cmc_scores (:py:class:`list`): A list in the format ``[(negatives,
        positives), ...]`` containing the CMC scores (i.e. :py:class:`list`:
        A list of tuples, where each tuple contains the
        ``negative`` and ``positive`` scores for one probe of the database).

        Each pair contains the ``negative`` and the ``positive`` scores for **one
        probe item**.  Each pair can contain up to one empty array (or ``None``),
        i.e., in case of open set recognition.

      threshold (float): The decision threshold :math:`\\tau``.


    Returns:

      float: The false alarm rate.

    """
    incorrect = 0
    counter = 0
    for neg, pos in cmc_scores:
        # we only consider the out-of-gallery probes, i.e., with no positive scores
        if pos is None or not numpy.array(pos).size:
            counter += 1

            # check if the probe is above threshold
            if neg is None or not numpy.array(neg).size:
                raise ValueError(
                    "One pair of the CMC scores has neither positive nor negative values"
                )
            if numpy.max(neg) >= threshold:
                incorrect += 1

    if not counter:
        logger.warn("No out-of-gallery probe was found")
        return 0.0

    return float(incorrect) / float(counter)


def eer(negatives, positives, is_sorted=False, also_farfrr=False):
    """Calculates the Equal Error Rate (EER).

    Please note that it is possible that eer != fpr != fnr.
    This function returns (fpr + fnr) / 2 as eer.
    If you also need the fpr and fnr values, set ``also_farfrr`` to ``True``.

    Parameters
    ----------
    negatives : ``array_like (1D, float)``
        The scores for comparisons of objects of different classes.
    positives : ``array_like (1D, float)``
        The scores for comparisons of objects of the same class.
    is_sorted : bool
        Are both sets of scores already in ascendantly sorted order?
    also_farfrr : bool
        If True, it will also return far and frr.

    Returns
    -------
    eer : float
        The Equal Error Rate (EER).
    fpr : float
        The False Positive Rate (FPR). Returned only when ``also_farfrr`` is
        ``True``.
    fnr : float
        The False Negative Rate (FNR). Returned only when ``also_farfrr`` is
        ``True``.
    """
    threshold = eer_threshold(negatives, positives, is_sorted)
    far, frr = farfrr(negatives, positives, threshold)
    if also_farfrr:
        return (far + frr) / 2.0, far, frr
    return (far + frr) / 2.0


def roc_auc_score(
    negatives, positives, npoints=2000, min_far=-8, log_scale=False
):
    """Area Under the ROC Curve.
    Computes the area under the ROC curve. This is useful when you want to report one
    number that represents an ROC curve. This implementation uses the trapezoidal rule for
    the integration of the ROC curve. For more information, see:
    https://en.wikipedia.org/wiki/Receiver_operating_characteristic#Area_under_the_curve


    Parameters
    ----------
    negatives : array_like
        The negative scores.
    positives : array_like
        The positive scores.
    npoints : int, optional
        Number of points in the ROC curve. Higher numbers leads to more accurate ROC.
    min_far : float, optional
        Min FAR and FRR values to consider when calculating ROC.
    log_scale : bool, optional
        If True, converts the x axis (FPR) to log10 scale before calculating AUC. This is
        useful in cases where len(negatives) >> len(positives)

    Returns
    -------
    float
        The ROC AUC. If ``log_scale`` is False, the value should be between 0 and 1.
    """
    fpr, fnr = roc(negatives, positives, npoints, min_far=min_far)
    tpr = 1 - fnr

    if log_scale:
        fpr_pos = fpr > 0
        fpr, tpr = fpr[fpr_pos], tpr[fpr_pos]
        fpr = numpy.log10(fpr)

    area = -1 * numpy.trapz(tpr, fpr)
    return area



#!/usr/bin/env python
# vim: set fileencoding=utf-8 :
# Thu May 16 11:41:49 CEST 2013

"""Measures for calibration"""

import math

import numpy


def cllr(negatives, positives):
    """Cost of log likelihood ratio as defined by the Bosaris toolkit

    Computes the 'cost of log likelihood ratio' (:math:`C_{llr}`) measure as
    given in the Bosaris toolkit


    Parameters:

      negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier.

      positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier.


    Returns:

      float: The computed :math:`C_{llr}` value.

    """
    sum_pos, sum_neg = 0.0, 0.0
    for pos in positives:
        sum_pos += math.log(1.0 + math.exp(-pos), 2.0)
    for neg in negatives:
        sum_neg += math.log(1.0 + math.exp(neg), 2.0)
    return (sum_pos / len(positives) + sum_neg / len(negatives)) / 2.0


def min_cllr(negatives, positives):
    """Minimum cost of log likelihood ratio as defined by the Bosaris toolkit

    Computes the 'minimum cost of log likelihood ratio' (:math:`C_{llr}^{min}`)
    measure as given in the bosaris toolkit


    Parameters:

      negatives (array): 1D float array that contains the scores of the
        "negative" (noise, non-class) samples of your classifier.

      positives (array): 1D float array that contains the scores of the
        "positive" (signal, class) samples of your classifier.


    Returns:

      float: The computed :math:`C_{llr}^{min}` value.

    """

    # first, sort both scores
    neg = sorted(negatives)
    pos = sorted(positives)
    N = len(neg)
    P = len(pos)
    II = N + P
    # now, iterate through both score sets and add a 0 for negative and 1 for
    # positive scores
    n, p = 0, 0
    ideal = numpy.zeros(II)
    neg_indices = [0] * N
    pos_indices = [0] * P
    for i in range(II):
        if p < P and (n == N or neg[n] > pos[p]):
            pos_indices[p] = i
            p += 1
            ideal[i] = 1
        else:
            neg_indices[n] = i
            n += 1

    # compute the pool adjacent violaters method on the ideal LLR scores
    ghat = numpy.ndarray(ideal.shape, dtype=numpy.float)
    raise NotImplementedError("No pavx implementation")
    pavx(ideal, ghat)  # noqa: F821

    # disable runtime warnings for a short time since log(0) will raise a warning
    old_warn_setup = numpy.seterr(divide="ignore")
    # ... compute logs
    posterior_log_odds = numpy.log(ghat) - numpy.log(1.0 - ghat)
    log_prior_odds = math.log(float(P) / float(N))
    # ... activate old warnings
    numpy.seterr(**old_warn_setup)

    llrs = posterior_log_odds - log_prior_odds

    # some weired addition
    #  for i in range(II):
    #    llrs[i] += float(i)*1e-6/float(II)

    # unmix positive and negative scores
    new_neg = numpy.zeros(N)
    for n in range(N):
        new_neg[n] = llrs[neg_indices[n]]
    new_pos = numpy.zeros(P)
    for p in range(P):
        new_pos[p] = llrs[pos_indices[p]]

    # compute cllr of these new 'optimal' LLR scores
    return cllr(new_neg, new_pos)



"""Runs error analysis on score sets, outputs metrics and plots"""

from __future__ import division, print_function

import logging
import math
import sys

from abc import ABCMeta, abstractmethod

import click
import matplotlib
import matplotlib.pyplot as mpl
import numpy

from matplotlib import gridspec
from matplotlib.backends.backend_pdf import PdfPages
from tabulate import tabulate

from .. import far_threshold, plot, ppndf, utils

LOGGER = logging.getLogger("bob.measure")


def check_list_value(values, desired_number, name, name2="systems"):
    if values is not None and len(values) != desired_number:
        if len(values) == 1:
            values = values * desired_number
        else:
            raise click.BadParameter(
                "#{} ({}) must be either 1 value or the same as "
                "#{} ({} values)".format(name, values, name2, desired_number)
            )

    return values


class MeasureBase(object):
    """Base class for metrics and plots.
    This abstract class define the framework to plot or compute metrics from a
    list of (positive, negative) scores tuples.

    Attributes
    ----------
    func_load:
        Function that is used to load the input files
    """

    __metaclass__ = ABCMeta  # for python 2.7 compatibility

    def __init__(self, ctx, scores, evaluation, func_load):
        """
        Parameters
        ----------
        ctx : :py:class:`dict`
            Click context dictionary.

        scores : :any:`list`:
            List of input files (e.g. dev-{1, 2, 3}, {dev,eval}-scores1
            {dev,eval}-scores2)
        eval : :py:class:`bool`
            True if eval data are used
        func_load : Function that is used to load the input files
        """
        self._scores = scores
        self._ctx = ctx
        self.func_load = func_load
        self._legends = ctx.meta.get("legends")
        self._eval = evaluation
        self._min_arg = ctx.meta.get("min_arg", 1)
        if len(scores) < 1 or len(scores) % self._min_arg != 0:
            raise click.BadParameter(
                "Number of argument must be a non-zero multiple of %d"
                % self._min_arg
            )
        self.n_systems = int(len(scores) / self._min_arg)
        if self._legends is not None and len(self._legends) < self.n_systems:
            raise click.BadParameter(
                "Number of legends must be >= to the " "number of systems"
            )

    def run(self):
        """Generate outputs (e.g. metrics, files, pdf plots).
        This function calls abstract methods
        :func:`~bob.measure.script.figure.MeasureBase.init_process` (before
        loop), :py:func:`~bob.measure.script.figure.MeasureBase.compute`
        (in the loop iterating through the different
        systems) and :py:func:`~bob.measure.script.figure.MeasureBase.end_process`
        (after the loop).
        """
        # init matplotlib, log files, ...
        self.init_process()
        # iterates through the different systems and feed `compute`
        # with the dev (and eval) scores of each system
        # Note that more than one dev or eval scores score can be passed to
        # each system
        for idx in range(self.n_systems):
            # load scores for each system: get the corresponding arrays and
            # base-name of files
            input_scores, input_names = self._load_files(
                # Scores are given as followed:
                # SysA-dev SysA-eval ... SysA-XX  SysB-dev SysB-eval ... SysB-XX
                # ------------------------------  ------------------------------
                #   First set of `self._min_arg`     Second set of input files
                #     input files starting at               for SysB
                #    index idx * self._min_arg
                self._scores[idx * self._min_arg : (idx + 1) * self._min_arg]
            )
            LOGGER.info("-----Input files for system %d-----", idx + 1)
            for i, name in enumerate(input_names):
                if not self._eval:
                    LOGGER.info("Dev. score %d: %s", i + 1, name)
                else:
                    if i % 2 == 0:
                        LOGGER.info("Dev. score %d: %s", i / 2 + 1, name)
                    else:
                        LOGGER.info("Eval. score %d: %s", i / 2 + 1, name)
            LOGGER.info("----------------------------------")

            self.compute(idx, input_scores, input_names)
        # setup final configuration, plotting properties, ...
        self.end_process()

    # protected functions that need to be overwritten
    def init_process(self):
        """Called in :py:func:`~bob.measure.script.figure.MeasureBase`.run
        before iterating through the different systems.
        Should reimplemented in derived classes"""
        pass

    # Main computations are done here in the subclasses
    @abstractmethod
    def compute(self, idx, input_scores, input_names):
        """Compute metrics or plots from the given scores provided by
        :py:func:`~bob.measure.script.figure.MeasureBase.run`.
        Should reimplemented in derived classes

        Parameters
        ----------
        idx : :obj:`int`
            index of the system
        input_scores: :any:`list`
            list of scores returned by the loading function
        input_names: :any:`list`
            list of base names for the input file of the system
        """
        pass
        # structure of input is (vuln example):
        # if evaluation is provided
        # [ (dev_licit_neg, dev_licit_pos), (eval_licit_neg, eval_licit_pos),
        #   (dev_spoof_neg, dev_licit_pos), (eval_spoof_neg, eval_licit_pos)]
        # and if only dev:
        # [ (dev_licit_neg, dev_licit_pos), (dev_spoof_neg, dev_licit_pos)]

    # Things to do after the main iterative computations are done
    @abstractmethod
    def end_process(self):
        """Called in :py:func:`~bob.measure.script.figure.MeasureBase`.run
        after iterating through the different systems.
        Should reimplemented in derived classes"""
        pass

    # common protected functions

    def _load_files(self, filepaths):
        """Load the input files and return the base names of the files

        Returns
        -------
            scores: :any:`list`:
                A list that contains the output of
                ``func_load`` for the given files
            basenames: :any:`list`:
                A list of the given files
        """
        scores = []
        basenames = []
        for filename in filepaths:
            basenames.append(filename)
            scores.append(self.func_load(filename))
        return scores, basenames


class Metrics(MeasureBase):
    """Compute metrics from score files

    Attributes
    ----------
    log_file: str
        output stream
    """

    def __init__(
        self,
        ctx,
        scores,
        evaluation,
        func_load,
        names=(
            "False Positive Rate",
            "False Negative Rate",
            "Precision",
            "Recall",
            "F1-score",
            "Area Under ROC Curve",
            "Area Under ROC Curve (log scale)",
        ),
    ):
        super(Metrics, self).__init__(ctx, scores, evaluation, func_load)
        self.names = names
        self._tablefmt = ctx.meta.get("tablefmt")
        self._criterion = ctx.meta.get("criterion")
        self._open_mode = ctx.meta.get("open_mode")
        self._thres = ctx.meta.get("thres")
        self._decimal = ctx.meta.get("decimal", 2)
        if self._thres is not None:
            if len(self._thres) == 1:
                self._thres = self._thres * self.n_systems
            elif len(self._thres) != self.n_systems:
                raise click.BadParameter(
                    "#thresholds must be the same as #systems (%d)"
                    % len(self.n_systems)
                )
        self._far = ctx.meta.get("far_value")
        self._log = ctx.meta.get("log")
        self.log_file = sys.stdout
        if self._log is not None:
            self.log_file = open(self._log, self._open_mode)

    def get_thres(self, criterion, dev_neg, dev_pos, far):
        return utils.get_thres(criterion, dev_neg, dev_pos, far)

    def _numbers(self, neg, pos, threshold, fta):
        from .. import f_score, farfrr, precision_recall, roc_auc_score

        # fpr and fnr
        fmr, fnmr = farfrr(neg, pos, threshold)
        hter = (fmr + fnmr) / 2.0
        far = fmr * (1 - fta)
        frr = fta + fnmr * (1 - fta)

        ni = neg.shape[0]  # number of impostors
        fm = int(round(fmr * ni))  # number of false accepts
        nc = pos.shape[0]  # number of clients
        fnm = int(round(fnmr * nc))  # number of false rejects

        # precision and recall
        precision, recall = precision_recall(neg, pos, threshold)

        # f_score
        f1_score = f_score(neg, pos, threshold, 1)

        # AUC ROC
        auc = roc_auc_score(neg, pos)
        auc_log = roc_auc_score(neg, pos, log_scale=True)
        return (
            fta,
            fmr,
            fnmr,
            hter,
            far,
            frr,
            fm,
            ni,
            fnm,
            nc,
            precision,
            recall,
            f1_score,
            auc,
            auc_log,
        )

    def _strings(self, metrics):
        n_dec = ".%df" % self._decimal
        fta_str = "%s%%" % format(100 * metrics[0], n_dec)
        fmr_str = "%s%% (%d/%d)" % (
            format(100 * metrics[1], n_dec),
            metrics[6],
            metrics[7],
        )
        fnmr_str = "%s%% (%d/%d)" % (
            format(100 * metrics[2], n_dec),
            metrics[8],
            metrics[9],
        )
        far_str = "%s%%" % format(100 * metrics[4], n_dec)
        frr_str = "%s%%" % format(100 * metrics[5], n_dec)
        hter_str = "%s%%" % format(100 * metrics[3], n_dec)
        prec_str = "%s" % format(metrics[10], n_dec)
        recall_str = "%s" % format(metrics[11], n_dec)
        f1_str = "%s" % format(metrics[12], n_dec)
        auc_str = "%s" % format(metrics[13], n_dec)
        auc_log_str = "%s" % format(metrics[14], n_dec)

        return (
            fta_str,
            fmr_str,
            fnmr_str,
            far_str,
            frr_str,
            hter_str,
            prec_str,
            recall_str,
            f1_str,
            auc_str,
            auc_log_str,
        )

    def _get_all_metrics(self, idx, input_scores, input_names):
        """Compute all metrics for dev and eval scores"""
        neg_list, pos_list, fta_list = utils.get_fta_list(input_scores)
        dev_neg, dev_pos, dev_fta = neg_list[0], pos_list[0], fta_list[0]
        dev_file = input_names[0]
        if self._eval:
            eval_neg, eval_pos, eval_fta = neg_list[1], pos_list[1], fta_list[1]

        threshold = (
            self.get_thres(self._criterion, dev_neg, dev_pos, self._far)
            if self._thres is None
            else self._thres[idx]
        )

        title = self._legends[idx] if self._legends is not None else None
        if self._thres is None:
            far_str = ""
            if self._criterion == "far" and self._far is not None:
                far_str = str(self._far)
            click.echo(
                "[Min. criterion: %s %s] Threshold on Development set `%s`: %e"
                % (
                    self._criterion.upper(),
                    far_str,
                    title or dev_file,
                    threshold,
                ),
                file=self.log_file,
            )
        else:
            click.echo(
                "[Min. criterion: user provided] Threshold on "
                "Development set `%s`: %e" % (dev_file or title, threshold),
                file=self.log_file,
            )

        res = []
        res.append(
            self._strings(self._numbers(dev_neg, dev_pos, threshold, dev_fta))
        )

        if self._eval:
            # computes statistics for the eval set based on the threshold a
            # priori
            res.append(
                self._strings(
                    self._numbers(eval_neg, eval_pos, threshold, eval_fta)
                )
            )
        else:
            res.append(None)

        return res

    def compute(self, idx, input_scores, input_names):
        """Compute metrics thresholds and tables (FPR, FNR, precision, recall,
        f1_score) for given system inputs"""
        dev_file = input_names[0]
        title = self._legends[idx] if self._legends is not None else None
        all_metrics = self._get_all_metrics(idx, input_scores, input_names)
        fta_dev = float(all_metrics[0][0].replace("%", ""))
        if fta_dev > 0.0:
            LOGGER.warn(
                "NaNs scores (%s) were found in %s amd removed",
                all_metrics[0][0],
                dev_file,
            )
        headers = [" " or title, "Development"]
        rows = [
            [self.names[0], all_metrics[0][1]],
            [self.names[1], all_metrics[0][2]],
            [self.names[2], all_metrics[0][6]],
            [self.names[3], all_metrics[0][7]],
            [self.names[4], all_metrics[0][8]],
            [self.names[5], all_metrics[0][9]],
            [self.names[6], all_metrics[0][10]],
        ]

        if self._eval:
            eval_file = input_names[1]
            fta_eval = float(all_metrics[1][0].replace("%", ""))
            if fta_eval > 0.0:
                LOGGER.warn(
                    "NaNs scores (%s) were found in %s and removed.",
                    all_metrics[1][0],
                    eval_file,
                )
            # computes statistics for the eval set based on the threshold a
            # priori
            headers.append("Evaluation")
            rows[0].append(all_metrics[1][1])
            rows[1].append(all_metrics[1][2])
            rows[2].append(all_metrics[1][6])
            rows[3].append(all_metrics[1][7])
            rows[4].append(all_metrics[1][8])
            rows[5].append(all_metrics[1][9])
            rows[6].append(all_metrics[1][10])

        click.echo(tabulate(rows, headers, self._tablefmt), file=self.log_file)

    def end_process(self):
        """Close log file if needed"""
        if self._log is not None:
            self.log_file.close()


class MultiMetrics(Metrics):
    """Computes average of metrics based on several protocols (cross
    validation)

    Attributes
    ----------
    log_file : str
        output stream
    names : tuple
        List of names for the metrics.
    """

    def __init__(
        self,
        ctx,
        scores,
        evaluation,
        func_load,
        names=(
            "NaNs Rate",
            "False Positive Rate",
            "False Negative Rate",
            "False Accept Rate",
            "False Reject Rate",
            "Half Total Error Rate",
        ),
    ):
        super(MultiMetrics, self).__init__(
            ctx, scores, evaluation, func_load, names=names
        )

        self.headers = ["Methods"] + list(self.names)
        if self._eval:
            self.headers.insert(1, self.names[5] + " (dev)")
        self.rows = []

    def _strings(self, metrics):
        (
            ftam,
            fmrm,
            fnmrm,
            hterm,
            farm,
            frrm,
            _,
            _,
            _,
            _,
            _,
            _,
            _,
        ) = metrics.mean(axis=0)
        ftas, fmrs, fnmrs, hters, fars, frrs, _, _, _, _, _, _, _ = metrics.std(
            axis=0
        )
        n_dec = ".%df" % self._decimal
        fta_str = "%s%% (%s%%)" % (
            format(100 * ftam, n_dec),
            format(100 * ftas, n_dec),
        )
        fmr_str = "%s%% (%s%%)" % (
            format(100 * fmrm, n_dec),
            format(100 * fmrs, n_dec),
        )
        fnmr_str = "%s%% (%s%%)" % (
            format(100 * fnmrm, n_dec),
            format(100 * fnmrs, n_dec),
        )
        far_str = "%s%% (%s%%)" % (
            format(100 * farm, n_dec),
            format(100 * fars, n_dec),
        )
        frr_str = "%s%% (%s%%)" % (
            format(100 * frrm, n_dec),
            format(100 * frrs, n_dec),
        )
        hter_str = "%s%% (%s%%)" % (
            format(100 * hterm, n_dec),
            format(100 * hters, n_dec),
        )
        return fta_str, fmr_str, fnmr_str, far_str, frr_str, hter_str

    def compute(self, idx, input_scores, input_names):
        """Computes the average of metrics over several protocols."""
        neg_list, pos_list, fta_list = utils.get_fta_list(input_scores)
        step = 2 if self._eval else 1
        self._dev_metrics = []
        self._thresholds = []
        for i in range(0, len(input_scores), step):
            neg, pos, fta = neg_list[i], pos_list[i], fta_list[i]
            threshold = (
                self.get_thres(self._criterion, neg, pos, self._far)
                if self._thres is None
                else self._thres[idx]
            )
            self._thresholds.append(threshold)
            self._dev_metrics.append(self._numbers(neg, pos, threshold, fta))
        self._dev_metrics = numpy.array(self._dev_metrics)

        if self._eval:
            self._eval_metrics = []
            for i in range(1, len(input_scores), step):
                neg, pos, fta = neg_list[i], pos_list[i], fta_list[i]
                threshold = self._thresholds[i // 2]
                self._eval_metrics.append(
                    self._numbers(neg, pos, threshold, fta)
                )
            self._eval_metrics = numpy.array(self._eval_metrics)

        title = self._legends[idx] if self._legends is not None else None

        fta_str, fmr_str, fnmr_str, far_str, frr_str, hter_str = self._strings(
            self._dev_metrics
        )

        if self._eval:
            self.rows.append([title, hter_str])
        else:
            self.rows.append(
                [title, fta_str, fmr_str, fnmr_str, far_str, frr_str, hter_str]
            )

        if self._eval:
            # computes statistics for the eval set based on the threshold a
            # priori
            (
                fta_str,
                fmr_str,
                fnmr_str,
                far_str,
                frr_str,
                hter_str,
            ) = self._strings(self._eval_metrics)

            self.rows[-1].extend(
                [fta_str, fmr_str, fnmr_str, far_str, frr_str, hter_str]
            )

    def end_process(self):
        click.echo(
            tabulate(self.rows, self.headers, self._tablefmt),
            file=self.log_file,
        )
        super(MultiMetrics, self).end_process()


class PlotBase(MeasureBase):
    """Base class for plots. Regroup several options and code
    shared by the different plots
    """

    def __init__(self, ctx, scores, evaluation, func_load):
        super(PlotBase, self).__init__(ctx, scores, evaluation, func_load)
        self._output = ctx.meta.get("output")
        self._points = ctx.meta.get("points", 2000)
        self._split = ctx.meta.get("split")
        self._axlim = ctx.meta.get("axlim")
        self._alpha = ctx.meta.get("alpha")
        self._disp_legend = ctx.meta.get("disp_legend", True)
        self._legend_loc = ctx.meta.get("legend_loc")
        self._min_dig = None
        if "min_far_value" in ctx.meta:
            self._min_dig = int(math.log10(ctx.meta["min_far_value"]))
        elif self._axlim is not None and self._axlim[0] is not None:
            self._min_dig = int(
                math.log10(self._axlim[0]) if self._axlim[0] != 0 else 0
            )
        self._clayout = ctx.meta.get("clayout")
        self._far_at = ctx.meta.get("lines_at")
        self._trans_far_val = self._far_at
        if self._far_at is not None:
            self._eval_points = {line: [] for line in self._far_at}
            self._lines_val = []
        self._print_fn = ctx.meta.get("show_fn", True)
        self._x_rotation = ctx.meta.get("x_rotation")
        if "style" in ctx.meta:
            mpl.style.use(ctx.meta["style"])
        self._nb_figs = 2 if self._eval and self._split else 1
        self._colors = utils.get_colors(self.n_systems)
        self._line_linestyles = ctx.meta.get("line_styles", False)
        self._linestyles = utils.get_linestyles(
            self.n_systems, self._line_linestyles
        )
        self._titles = ctx.meta.get("titles", []) * 2
        # for compatibility
        self._title = ctx.meta.get("title")
        if not self._titles and self._title is not None:
            self._titles = [self._title] * 2

        self._x_label = ctx.meta.get("x_label")
        self._y_label = ctx.meta.get("y_label")
        self._grid_color = "silver"
        self._pdf_page = None
        self._end_setup_plot = True

    def init_process(self):
        """Open pdf and set axis font size if provided"""
        if not hasattr(matplotlib, "backends"):
            matplotlib.use("pdf")

        self._pdf_page = (
            self._ctx.meta["PdfPages"]
            if "PdfPages" in self._ctx.meta
            else PdfPages(self._output)
        )

        for i in range(self._nb_figs):
            fs = self._ctx.meta.get("figsize")
            fig = mpl.figure(i + 1, figsize=fs)
            fig.set_constrained_layout(self._clayout)
            fig.clear()

    def end_process(self):
        """Set title, legend, axis labels, grid colors, save figures, drow
        lines and close pdf if needed"""
        # draw vertical lines
        if self._far_at is not None:
            for line, line_trans in zip(self._far_at, self._trans_far_val):
                mpl.figure(1)
                mpl.plot(
                    [line_trans, line_trans],
                    [-100.0, 100.0],
                    "--",
                    color="black",
                )
                if self._eval and self._split:
                    mpl.figure(2)
                    x_values = [i for i, _ in self._eval_points[line]]
                    y_values = [j for _, j in self._eval_points[line]]
                    sort_indice = sorted(
                        range(len(x_values)), key=x_values.__getitem__
                    )
                    x_values = [x_values[i] for i in sort_indice]
                    y_values = [y_values[i] for i in sort_indice]
                    mpl.plot(x_values, y_values, "--", color="black")
        # only for plots
        if self._end_setup_plot:
            for i in range(self._nb_figs):
                fig = mpl.figure(i + 1)
                title = "" if not self._titles else self._titles[i]
                mpl.title(title if title.replace(" ", "") else "")
                mpl.xlabel(self._x_label)
                mpl.ylabel(self._y_label)
                mpl.grid(True, color=self._grid_color)
                if self._disp_legend:
                    self.plot_legends()
                self._set_axis()
                mpl.xticks(rotation=self._x_rotation)
                self._pdf_page.savefig(fig)

        # do not want to close PDF when running evaluate
        if "PdfPages" in self._ctx.meta and (
            "closef" not in self._ctx.meta or self._ctx.meta["closef"]
        ):
            self._pdf_page.close()

    def plot_legends(self):
        """Print legend on current plot"""
        if not self._disp_legend:
            return

        lines = []
        labels = []
        for ax in mpl.gcf().get_axes():
            ali, ala = ax.get_legend_handles_labels()
            # avoid duplicates in legend
            for li, la in zip(ali, ala):
                if la not in labels:
                    lines.append(li)
                    labels.append(la)

        # create legend on the top or bottom axis
        leg = mpl.legend(
            lines,
            labels,
            loc=self._legend_loc,
            ncol=1,
        )

        return leg

    # common protected functions

    def _label(self, base, idx):
        if self._legends is not None and len(self._legends) > idx:
            return self._legends[idx]
        if self.n_systems > 1:
            return base + (" %d" % (idx + 1))
        return base

    def _set_axis(self):
        if self._axlim is not None:
            mpl.axis(self._axlim)


class Roc(PlotBase):
    """Handles the plotting of ROC"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Roc, self).__init__(ctx, scores, evaluation, func_load)
        self._titles = self._titles or ["ROC dev.", "ROC eval."]
        self._x_label = self._x_label or "FPR"
        self._semilogx = ctx.meta.get("semilogx", True)
        self._tpr = ctx.meta.get("tpr", True)
        dflt_y_label = "TPR" if self._tpr else "FNR"
        self._y_label = self._y_label or dflt_y_label
        best_legend = "lower right" if self._semilogx else "upper right"
        self._legend_loc = self._legend_loc or best_legend
        # custom defaults
        if self._axlim is None:
            self._axlim = [None, None, -0.05, 1.05]
        self._min_dig = -4 if self._min_dig is None else self._min_dig

    def compute(self, idx, input_scores, input_names):
        """Plot ROC for dev and eval data using
        :py:func:`bob.measure.plot.roc`"""
        neg_list, pos_list, _ = utils.get_fta_list(input_scores)
        dev_neg, dev_pos = neg_list[0], pos_list[0]
        dev_file = input_names[0]
        if self._eval:
            eval_neg, eval_pos = neg_list[1], pos_list[1]
            eval_file = input_names[1]

        mpl.figure(1)
        if self._eval:
            LOGGER.info("ROC dev. curve using %s", dev_file)
            plot.roc(
                dev_neg,
                dev_pos,
                npoints=self._points,
                semilogx=self._semilogx,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev", idx),
                alpha=self._alpha,
            )
            if self._split:
                mpl.figure(2)

            linestyle = "--" if not self._split else self._linestyles[idx]
            LOGGER.info("ROC eval. curve using %s", eval_file)
            plot.roc(
                eval_neg,
                eval_pos,
                linestyle=linestyle,
                npoints=self._points,
                semilogx=self._semilogx,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                label=self._label("eval.", idx),
                alpha=self._alpha,
            )
            if self._far_at is not None:
                from .. import fprfnr

                for line in self._far_at:
                    thres_line = far_threshold(dev_neg, dev_pos, line)
                    eval_fmr, eval_fnmr = fprfnr(eval_neg, eval_pos, thres_line)
                    if self._tpr:
                        eval_fnmr = 1 - eval_fnmr
                    mpl.scatter(eval_fmr, eval_fnmr, c=self._colors[idx], s=30)
                    self._eval_points[line].append((eval_fmr, eval_fnmr))
        else:
            LOGGER.info("ROC dev. curve using %s", dev_file)
            plot.roc(
                dev_neg,
                dev_pos,
                npoints=self._points,
                semilogx=self._semilogx,
                tpr=self._tpr,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev", idx),
                alpha=self._alpha,
            )


class Det(PlotBase):
    """Handles the plotting of DET"""

    def __init__(self, ctx, scores, evaluation, func_load):
        super(Det, self).__init__(ctx, scores, evaluation, func_load)
        self._titles = self._titles or ["DET dev.", "DET eval."]
        self._x_label = self._x_label or "FPR (%)"
        self._y_label = self._y_label or "FNR (%)"
        self._legend_loc = self._legend_loc or "upper right"
        if self._far_at is not None:
            self._trans_far_val = ppndf(self._far_at)
        # custom defaults here
        if self._x_rotation is None:
            self._x_rotation = 50

        if self._axlim is None:
            self._axlim = [0.01, 99, 0.01, 99]

        if self._min_dig is not None:
            self._axlim[0] = math.pow(10, self._min_dig) * 100

        self._min_dig = -4 if self._min_dig is None else self._min_dig

    def compute(self, idx, input_scores, input_names):
        """Plot DET for dev and eval data using
        :py:func:`bob.measure.plot.det`"""
        neg_list, pos_list, _ = utils.get_fta_list(input_scores)
        dev_neg, dev_pos = neg_list[0], pos_list[0]
        dev_file = input_names[0]
        if self._eval:
            eval_neg, eval_pos = neg_list[1], pos_list[1]
            eval_file = input_names[1]

        mpl.figure(1)
        if self._eval and eval_neg is not None:
            LOGGER.info("DET dev. curve using %s", dev_file)
            plot.det(
                dev_neg,
                dev_pos,
                self._points,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev.", idx),
                alpha=self._alpha,
            )
            if self._split:
                mpl.figure(2)
            linestyle = "--" if not self._split else self._linestyles[idx]
            LOGGER.info("DET eval. curve using %s", eval_file)
            plot.det(
                eval_neg,
                eval_pos,
                self._points,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=linestyle,
                label=self._label("eval.", idx),
                alpha=self._alpha,
            )
            if self._far_at is not None:
                from .. import farfrr

                for line in self._far_at:
                    thres_line = far_threshold(dev_neg, dev_pos, line)
                    eval_fmr, eval_fnmr = farfrr(eval_neg, eval_pos, thres_line)
                    eval_fmr, eval_fnmr = ppndf(eval_fmr), ppndf(eval_fnmr)
                    mpl.scatter(eval_fmr, eval_fnmr, c=self._colors[idx], s=30)
                    self._eval_points[line].append((eval_fmr, eval_fnmr))
        else:
            LOGGER.info("DET dev. curve using %s", dev_file)
            plot.det(
                dev_neg,
                dev_pos,
                self._points,
                min_far=self._min_dig,
                color=self._colors[idx],
                linestyle=self._linestyles[idx],
                label=self._label("dev.", idx),
                alpha=self._alpha,
            )

    def _set_axis(self):
        plot.det_axis(self._axlim)


class Epc(PlotBase):
    """Handles the plotting of EPC"""

    def __init__(self, ctx, scores, evaluation, func_load, hter="HTER"):
        super(Epc, self).__init__(ctx, scores, evaluation, func_load)
        if self._min_arg != 2:
            raise click.UsageError("EPC requires dev. and eval. score files")
        self._titles = self._titles or ["EPC"] * 2
        self._x_label = self._x_label or r"$\alpha$"
        self._y_label = self._y_label or hter + " (%)"
        self._legend_loc = self._legend_loc or "upper center"
        self._eval = True  # always eval data with EPC
        self._split = False
        self._nb_figs = 1
        self._far_at = None

    def compute(self, idx, input_scores, input_names):
        """Plot EPC using :py:func:`bob.measure.plot.epc`"""
        neg_list, pos_list, _ = utils.get_fta_list(input_scores)
        dev_neg, dev_pos = neg_list[0], pos_list[0]
        dev_file = input_names[0]
        if self._eval:
            eval_neg, eval_pos = neg_list[1], pos_list[1]
            eval_file = input_names[1]

        LOGGER.info("EPC using %s", dev_file + "_" + eval_file)
        plot.epc(
            dev_neg,
            dev_pos,
            eval_neg,
            eval_pos,
            self._points,
            color=self._colors[idx],
            linestyle=self._linestyles[idx],
            label=self._label("curve", idx),
            alpha=self._alpha,
        )


class GridSubplot(PlotBase):
    """A base class for plots that contain subplots and legends.

    To use this class, use `create_subplot` in `compute` each time you need a
    new axis. and call `finalize_one_page` in `compute` when a page is finished
    rendering.
    """

    def __init__(self, ctx, scores, evaluation, func_load):
        super(GridSubplot, self).__init__(ctx, scores, evaluation, func_load)

        # Check legend
        self._legend_loc = self._legend_loc or "upper center"
        if self._legend_loc == "best":
            self._legend_loc = "upper center"
        if "upper" not in self._legend_loc and "lower" not in self._legend_loc:
            raise ValueError(
                "Only best, upper-*, and lower-* legend locations are supported!"
            )
        self._nlegends = ctx.meta.get("legends_ncol", 3)

        # subplot grid
        self._nrows = ctx.meta.get("n_row", 1)
        self._ncols = ctx.meta.get("n_col", 1)

    def init_process(self):
        super(GridSubplot, self).init_process()
        self._create_grid_spec()

    def _create_grid_spec(self):
        # create a compatible GridSpec
        self._gs = gridspec.GridSpec(
            self._nrows,
            self._ncols,
            figure=mpl.gcf(),
        )

    def create_subplot(self, n, shared_axis=None):
        i, j = numpy.unravel_index(n, (self._nrows, self._ncols))
        axis = mpl.gcf().add_subplot(
            self._gs[i : i + 1, j : j + 1], sharex=shared_axis
        )
        return axis

    def finalize_one_page(self):
        # print legend on the page
        self.plot_legends()
        fig = mpl.gcf()
        axes = fig.get_axes()

        LOGGER.debug("%s contains %d axes:", fig, len(axes))
        for i, ax in enumerate(axes, start=1):
            LOGGER.debug("Axes %d: %s", i, ax)

        self._pdf_page.savefig(bbox_inches="tight")
        mpl.clf()
        mpl.figure()
        self._create_grid_spec()

    def plot_legends(self):
        """Print legend on current page"""
        if not self._disp_legend:
            return

        lines = []
        labels = []
        for ax in mpl.gcf().get_axes():
            ali, ala = ax.get_legend_handles_labels()
            # avoid duplicates in legend
            for li, la in zip(ali, ala):
                if la not in labels:
                    lines.append(li)
                    labels.append(la)

        # create legend on the top or bottom axis
        fig = mpl.gcf()
        if "upper" in self._legend_loc:
            # Set anchor to top of figure
            bbox_to_anchor = (0.0, 1.0, 1.0, 0.0)
            # Legend will be anchored with its bottom side, so switch the loc
            anchored_loc = self._legend_loc.replace("upper", "lower")
        else:
            # Set anchor to bottom of figure
            bbox_to_anchor = (0.0, 0.0, 1.0, 0.0)
            # Legend will be anchored with its top side, so switch the loc
            anchored_loc = self._legend_loc.replace("lower", "upper")
        leg = fig.legend(
            lines,
            labels,
            loc=anchored_loc,
            ncol=self._nlegends,
            bbox_to_anchor=bbox_to_anchor,
        )

        return leg


class Hist(GridSubplot):
    """Functional base class for histograms"""

    def __init__(self, ctx, scores, evaluation, func_load, nhist_per_system=2):
        super(Hist, self).__init__(ctx, scores, evaluation, func_load)
        self._nbins = ctx.meta.get("n_bins", ["doane"])
        self._nhist_per_system = nhist_per_system
        self._nbins = check_list_value(
            self._nbins, nhist_per_system, "n_bins", "histograms"
        )
        self._thres = ctx.meta.get("thres")
        self._thres = check_list_value(
            self._thres, self.n_systems, "thresholds"
        )
        self._criterion = ctx.meta.get("criterion")
        # no vertical (threshold) is displayed
        self._no_line = ctx.meta.get("no_line", False)
        # do not display dev histo
        self._hide_dev = ctx.meta.get("hide_dev", False)
        if self._hide_dev and not self._eval:
            raise click.BadParameter(
                "You can only use --hide-dev along with --eval"
            )
        # dev hist are displayed next to eval hist
        self._nrows *= 1 if self._hide_dev or not self._eval else 2
        self._nlegends = ctx.meta.get("legends_ncol", 3)

        # number of subplot on one page
        self._step_print = int(self._nrows * self._ncols)
        self._title_base = "Scores"
        self._y_label = self._y_label or "Probability density"
        self._x_label = self._x_label or "Score values"
        self._end_setup_plot = False
        # overide _titles of PlotBase
        self._titles = ctx.meta.get("titles", []) * 2

    def compute(self, idx, input_scores, input_names):
        """Draw histograms of negative and positive scores."""
        (
            dev_neg,
            dev_pos,
            eval_neg,
            eval_pos,
            threshold,
        ) = self._get_neg_pos_thres(idx, input_scores, input_names)

        # keep id of the current system
        sys = idx
        # if the id of the current system does not match the id of the plot,
        # change it
        if not self._hide_dev and self._eval:
            row = int(idx / self._ncols) * 2
            col = idx % self._ncols
            idx = col + self._ncols * row

        dev_axis = None

        if not self._hide_dev or not self._eval:
            dev_axis = self._print_subplot(
                idx,
                sys,
                dev_neg,
                dev_pos,
                threshold,
                not self._no_line,
                False,
            )

        if self._eval:
            idx += self._ncols if not self._hide_dev else 0
            self._print_subplot(
                idx,
                sys,
                eval_neg,
                eval_pos,
                threshold,
                not self._no_line,
                True,
                shared_axis=dev_axis,
            )

    def _print_subplot(
        self,
        idx,
        sys,
        neg,
        pos,
        threshold,
        draw_line,
        evaluation,
        shared_axis=None,
    ):
        """print a subplot for the given score and subplot index"""
        n = idx % self._step_print
        col = n % self._ncols
        sub_plot_idx = n + 1
        axis = self.create_subplot(n, shared_axis)
        self._setup_hist(neg, pos)
        if col == 0:
            axis.set_ylabel(self._y_label)
        # systems per page
        sys_per_page = self._step_print / (
            1 if self._hide_dev or not self._eval else 2
        )
        # rest to be printed
        sys_idx = sys % sys_per_page
        rest_print = self.n_systems - int(sys / sys_per_page) * sys_per_page
        # lower histo only
        is_lower = evaluation or not self._eval
        if is_lower and sys_idx + self._ncols >= min(sys_per_page, rest_print):
            axis.set_xlabel(self._x_label)
        dflt_title = "Eval. scores" if evaluation else "Dev. scores"
        if self.n_systems == 1 and (not self._eval or self._hide_dev):
            dflt_title = " "
        add = self.n_systems if is_lower else 0
        axis.set_title(self._get_title(sys + add, dflt_title))
        label = "%s threshold%s" % (
            "" if self._criterion is None else self._criterion.upper(),
            " (dev)" if self._eval else "",
        )
        if draw_line:
            self._lines(threshold, label, neg, pos, idx)

        # enable the grid and set it below other elements
        axis.set_axisbelow(True)
        axis.grid(True, color=self._grid_color)

        # if it was the last subplot of the page or the last subplot
        # to display, save figure
        if self._step_print == sub_plot_idx or (
            is_lower and sys == self.n_systems - 1
        ):
            self.finalize_one_page()
        return axis

    def _get_title(self, idx, dflt=None):
        """Get the histo title for the given idx"""
        title = (
            self._titles[idx]
            if self._titles is not None and idx < len(self._titles)
            else dflt
        )
        title = title or self._title_base
        title = (
            "" if title is not None and not title.replace(" ", "") else title
        )
        return title or ""

    def _get_neg_pos_thres(self, idx, input_scores, input_names):
        """Get scores and threshod for the given system at index idx"""
        neg_list, pos_list, _ = utils.get_fta_list(input_scores)
        length = len(neg_list)
        # lists returned by get_fta_list contains all the following items:
        # for bio or measure without eval:
        #   [dev]
        # for vuln with {licit,spoof} with eval:
        #   [dev, eval]
        # for vuln with {licit,spoof} without eval:
        #   [licit_dev, spoof_dev]
        # for vuln with {licit,spoof} with eval:
        #   [licit_dev, licit_eval, spoof_dev, spoof_eval]
        step = 2 if self._eval else 1
        # can have several files for one system
        dev_neg = [neg_list[x] for x in range(0, length, step)]
        dev_pos = [pos_list[x] for x in range(0, length, step)]
        eval_neg = eval_pos = None
        if self._eval:
            eval_neg = [neg_list[x] for x in range(1, length, step)]
            eval_pos = [pos_list[x] for x in range(1, length, step)]

        threshold = (
            utils.get_thres(self._criterion, dev_neg[0], dev_pos[0])
            if self._thres is None
            else self._thres[idx]
        )
        return dev_neg, dev_pos, eval_neg, eval_pos, threshold

    def _density_hist(self, scores, n, **kwargs):
        """Plots one density histo"""
        n, bins, patches = mpl.hist(
            scores, density=True, bins=self._nbins[n], **kwargs
        )
        return (n, bins, patches)

    def _lines(
        self, threshold, label=None, neg=None, pos=None, idx=None, **kwargs
    ):
        """Plots vertical line at threshold"""
        label = label or "Threshold"
        kwargs.setdefault("color", "C3")
        kwargs.setdefault("linestyle", "--")
        kwargs.setdefault("label", label)
        # plot a vertical threshold line
        mpl.axvline(x=threshold, ymin=0, ymax=1, **kwargs)

    def _setup_hist(self, neg, pos):
        """This function can be overwritten in derived classes

        Plots all the density histo required in one plot. Here negative and
        positive scores densities.
        """
        self._density_hist(
            neg[0], n=0, label="Negatives", alpha=0.5, color="C3"
        )
        self._density_hist(
            pos[0], n=1, label="Positives", alpha=0.5, color="C0"
        )



#!/usr/bin/env python
# coding=utf-8


"""Generate random scores.
"""

import logging
import os

import click
import numpy
import numpy.random

from clapper.click import verbosity_option
from click.types import FLOAT

logger = logging.getLogger(__name__)

NUM_NEG = 5000
NUM_POS = 5000


def gen_score_distr(mean_neg, mean_pos, sigma_neg=1, sigma_pos=1):
    """Generate scores from normal distributions

    Parameters
    ----------

    mean_neg : float
        Mean for negative scores

    mean_pos : float
        Mean for positive scores

    sigma_neg : float
        STDev for negative scores

    sigma_pos : float
        STDev for positive scores


    Returns
    -------

    neg_scores : numpy.ndarray (1D, float)
        Negatives scores

    pos_scores : numpy.ndarray (1D, float)
        Positive scores

    """

    return numpy.random.normal(
        mean_neg, sigma_neg, (NUM_NEG,)
    ), numpy.random.normal(mean_pos, sigma_pos, (NUM_POS,))


def write_scores_to_file(neg, pos, filename):
    """Writes score distributions into 2-column score files.

    For the format of the 2-column score files, please refer to Bob's
    documentation. See :py:func:`bob.measure.load.split`.

    Parameters
    ----------

    neg : :py:class:`numpy.ndarray`
        Scores for negative samples.

    pos : :py:class:`numpy.ndarray`
        Scores for positive samples.

    filename : str
        The path to write the score to.

    """

    os.makedirs(os.path.dirname(filename), exist_ok=True)

    with open(filename, "wt") as f:
        for i in pos:
            text = (
                "1 %f\n" % i if numpy.random.normal(0, 1) > 0.01 else "1 nan\n"
            )
            f.write(text)

        for i in neg:
            text = (
                "-1 %f\n" % i if numpy.random.normal(0, 1) > 0.01 else "1 nan\n"
            )
            f.write(text)


@click.command()
@click.argument("outdir")
@click.option("--mean-neg", default=-1, type=FLOAT, show_default=True)
@click.option("--mean-pos", default=1, type=FLOAT, show_default=True)
@verbosity_option(logger, expose_value=False)
def gen(outdir, mean_neg, mean_pos):
    """Generate random scores.

    Generates random scores for negative and positive scores, whatever they
    could be. The scores are generated using Gaussian distribution whose mean
    is an input parameter. The generated scores can be used as hypothetical
    datasets.
    """
    # Generate the data
    neg_dev, pos_dev = gen_score_distr(mean_neg, mean_pos)
    neg_eval, pos_eval = gen_score_distr(mean_neg, mean_pos)

    # Write the data into files
    write_scores_to_file(neg_dev, pos_dev, os.path.join(outdir, "scores-dev"))
    write_scores_to_file(
        neg_eval, pos_eval, os.path.join(outdir, "scores-eval")
    )



"""Tests for bob.measure scripts"""

import os

import click
import pkg_resources

from click.testing import CliRunner

from bob.io.base.testing_utils import assert_click_runner_result
from bob.measure.script import commands


def _F(f):
    """Returns the name of a file in the "data" subdirectory"""
    return pkg_resources.resource_filename(__name__, os.path.join("data", f))


def test_metrics():
    dev1 = _F("dev-1.txt")
    runner = CliRunner()
    result = runner.invoke(commands.metrics, [dev1])
    with runner.isolated_filesystem():
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)

    dev2 = _F("dev-2.txt")
    test1 = _F("test-1.txt")
    test2 = _F("test-2.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics, ["-e", dev1, test1, dev2, test2]
        )
        with open("tmp", "w") as f:
            f.write(result.output)
        assert_click_runner_result(result)
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.metrics,
            ["-e", "-l", "tmp", dev1, test1, dev2, test2, "-lg", "A,B"],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(commands.metrics, ["-l", "tmp", dev1, dev2])
        assert_click_runner_result(result)


def test_roc():
    dev1 = _F("dev-1.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.roc, ["--output", "test.pdf", dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    dev2 = _F("dev-2.txt")
    test1 = _F("test-1.txt")
    test2 = _F("test-2.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.roc,
            [
                "--split",
                "--output",
                "test.pdf",
                "-e",
                "-ts",
                "A,",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.roc,
            [
                "-e",
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_det():
    dev1 = _F("dev-1.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.det, [dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)
    dev2 = _F("dev-2.txt")
    test1 = _F("test-1.txt")
    test2 = _F("test-2.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.det,
            [
                "-e",
                "--split",
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "-ll",
                "upper-right",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.det,
            ["--output", "test.pdf", "-e", dev1, test1, dev2, test2],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_epc():
    dev1 = _F("dev-1.txt")
    test1 = _F("test-1.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.epc, [dev1, test1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    dev2 = _F("dev-2.txt")
    test2 = _F("test-2.txt")
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.epc,
            [
                "--output",
                "test.pdf",
                "--legends",
                "A,B",
                "--titles",
                "TA,TB",
                "-ll",
                "upper-right",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_hist():
    dev1 = _F("dev-1.txt")
    test1 = _F("test-1.txt")
    dev2 = _F("dev-2.txt")
    test2 = _F("test-2.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.hist, [dev1])
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            [
                "--criterion",
                "min-hter",
                "--no-line",
                "--output",
                "HISTO.pdf",
                "-b",
                "30,100",
                dev1,
                dev2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            [
                "-e",
                "--criterion",
                "eer",
                "--output",
                "HISTO.pdf",
                "-b",
                "30,20",
                "-sp",
                111,
                "-ts",
                "A,B",
                dev1,
                test1,
                dev2,
                test2,
            ],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_hist_legends():
    dev1 = _F("dev-1.txt")
    test1 = _F("test-1.txt")
    dev2 = _F("dev-2.txt")
    test2 = _F("test-2.txt")
    runner = CliRunner()

    # share same title for dev/eval of each system
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            ["-e", "-sp", 111, "-ts", "A,B", dev1, test1, dev2, test2],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)

    # individual titles for dev and eval
    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.hist,
            ["-e", "-sp", 121, "-ts", "A,B,C,D", dev1, test1, dev2, test2],
        )
        if result.output:
            click.echo(result.output)
        assert_click_runner_result(result)


def test_evaluate():
    dev1 = _F("dev-1.txt")
    test1 = _F("test-1.txt")
    dev2 = _F("dev-2.txt")
    test2 = _F("test-2.txt")
    runner = CliRunner()
    with runner.isolated_filesystem():
        result = runner.invoke(commands.evaluate, [dev1])
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.evaluate,
            ["--output", "my_plots.pdf", "-n", 300, dev1, dev2],
        )
        assert_click_runner_result(result)

    with runner.isolated_filesystem():
        result = runner.invoke(
            commands.evaluate, ["-e", dev1, test1, dev2, test2]
        )
        assert_click_runner_result(result)
